#!/usr/bin/env python

#
# The Personis system is copyright 2000-2012 University of Sydney
#       Bob.Kummerfeld@Sydney.edu.au

# This file is part of Personis.

# Personis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Personis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Personis.  If not, see <http://www.gnu.org/licenses/>.


"""
model database stored in ".model" in the model directory::

    Identifier  modelname
    Description     of the model
    owner       owner of the model
    password    md5 hash of owner password for this model
    apps        dictionary of app names and keys
        appname     dictionary of info about that app
            password    md5 hash of the apps access key (password)
            description of the app

context database stored in ".context" in each context directory::

    Identifier  context name
    Description     of the context
    perms       a dictionary of permissions for apps
        appname     dictionary of permission info for that app
            permname    True/False value indicates permname permission for the app
                eg {"ask":False, "tell":True} can tell values but not ask
            "resolvers" list of resolvers that the app is allowed to use
                eg {"resolvers":["last1", "goal"]}
"""

import os, shelve, sys, fcntl, time, string, glob
import hashlib
from . import filedict
from types import *
from . import resolvers
from . import ev_filters
import json
from . import util
import logging
import traceback
from ..client import Evidence
from ..client import Context
from ..client import View
from ..client import Component as clientComponent
from ..client import Access as clientAccess

PERSONIS_VERSION = "$LastChangedRevision: 727 $"



def MkModel(model=None, modeldir=None, user=None, description=None):
    """make a model with name "model" in directory modeldir for "user" with "description"

    :param model:
    :param modeldir:
    :param user:
    :param description:
    :raise:
    """

    if modeldir == None:
        raise ValueError("modeldir is None in MkModel")
    modeldir = os.path.join(modeldir, model)
    os.mkdir(modeldir)
    mod, mod_shelf_fd = shelf_open(os.path.join(modeldir,".model"), "c")
    mod['Identifier'] = model
    mod['owner'] = user
    #p = hashlib.md5()
    #if password == None:
    #   password = ''
    #p.update(password)
    mod['password'] = '' #p.hexdigest()
    mod['apps'] = {} # dictionary of app names and keys
    mod['Description'] = description
    shelf_close(mod, mod_shelf_fd)
    con, con_shelf_fd = shelf_open(os.path.join(modeldir,".context"), "c")
    con['Identifier'] = '.'
    con['Description'] = 'root context'
    con['perms'] = {}
    shelf_close(con, con_shelf_fd)
    comps, comps_shelf_fd = shelf_open(os.path.join(modeldir,".components"), "c")
    shelf_close(comps, comps_shelf_fd)
    evdb,evdb_fd = shelf_open(os.path.join(modeldir,".evidence"), "c")
    shelf_close(evdb, evdb_fd)
    views, views_shelf_fd = shelf_open(os.path.join(modeldir,".views"), "c")
    shelf_close(views, views_shelf_fd)
    subs, subs_shelf_fd = shelf_open(os.path.join(modeldir,".subscriptions"), "c")
    shelf_close(subs, subs_shelf_fd)

def shelf_open(shelf_name, mode):
    db = filedict.FileDict(filename=shelf_name)
    return db, None
 
def shelf_close(theShelf, shelf_fd):
    pass
    #theShelf.close()

class Component(clientComponent):

    def filterevidence(self, model=None, context=[], resolver_args=None):
        """

        :param model:
        :param context:
        :param resolver_args:

        :return:
            if evidence_filter is "all": 
                returns all evidence objects for this component in a list
            if evidence_filter is "last10":
                return only the last 10 elements
            if evidence_filter is "last1":
                return only the last element
            if evidence_filter is None:
                return empty list
            else returns a list of evidence objects for which evidence_filter(evidence)
                return true

        """
        strId = self.Identifier
        if resolver_args == None:
            raise ValueError('no evidence filter')
        evidence_filter = resolver_args.get('evidence_filter')
        if evidence_filter == None:
            evidence_filter = 'all'
        if evidence_filter in model.evidencefilterlist:
            efilter = model.evidencefilterlist[evidence_filter]
        else:
            raise ValueError('unknown evidence filter "%s"'%(repr(evidence_filter)))

        try:
            evdb,evdb_fd = shelf_open(model._getcontextdir(context)+"/.evidence", "r")
        except:
            raise ValueError("tell: no evidence db for %s"%(repr(context)))
        
        evidence_list = []
        for k in evdb:
            if k.startswith(strId+':'):
                evidence_list += [evdb[k].__dict__]
        evidence_list = efilter(elist=evidence_list, arguments=resolver_args)

        shelf_close(evdb, evdb_fd)
        return evidence_list

    def getevidence(self, model=None, context=[], count=None):
        """if count is None, returns (last evidence obj, count)
            else returns (count evidence obj, count)
        """
        strId = self.Identifier
        try:
            evdb,evdb_fd = shelf_open(model._getcontextdir(context)+"/.evidence", "r")
        except:
            raise ValueError("tell: no evidence db for %s"%(repr(context)))
        if strId not in evdb:
            shelf_close(evdb, evdb_fd)
            return (None, None) # no evidence for this component
        if count == None:
            count = evdb[strId]
        if "%s:%d"%(strId,count) in evdb:
            evidence = evdb["%s:%d"%(strId,count)]
        else:
            evidence = None
        shelf_close(evdb, evdb_fd)
        return (evidence, count)

    def findevidence(self, model=None, context=[], evidence_time=None):
        """

        :param model:
        :param context:
        :param evidence_time:
        :return:
        :raise:
        """
        if evidence_time == None:
            raise ValueError("findevidence: no evidence time specified")
        try:
            evdb,evdb_fd = shelf_open(model._getcontextdir(context)+"/.evidence", "r")
        except:
            raise ValueError("tell: no evidence db for %s"%(repr(context)))
        if self.Identifier not in evdb:
            shelf_close(evdb, evdb_fd)
            return (None, None) # no evidence for this component
        count = evdb[self.Identifier]
        evidence = None
        while count > 0:
            if "%s:%d"%(self.Identifier,count) in evdb:
                evidence = evdb["%s:%d"%(self.Identifier,count)]
                break
            else:
                count = count-1
        if evidence == None:
            # no evidence found
            shelf_close(evdb, evdb_fd)
            return (None, None)

        lo = 0; hi = count
        while lo < hi:
            mid = (lo+hi)//2
            while mid > 0:
                if "%s:%d"%(self.Identifier,mid) in evdb:
                    evidence = evdb["%s:%d"%(self.Identifier,mid)]
                    break
                else:
                    mid = mid-1
            if mid == 0: # no more evidence found!
                shelf_close(evdb, evdb_fd)
                return (None, None)
                
            if evidence_time < evidence.time:
                hi = mid
            else:
                lo = mid+1
        # index of required evidence item is "lo"
        shelf_close(evdb, evdb_fd)
        return evidence,count


class Access(resolvers.Access, ev_filters.Access):
    """
    :param model: model name
    :param modeldir: model directory
    :param user: user name
    :param debug: password string
    :return: a user model access object 
    :raise:
    """

    def __init__(self, model=None, modeldir="", user='', debug=0):
        if model == None:
            raise ValueError("model is None")
        if modeldir[0] != '/':
            self.modeldir = os.getcwd()+'/'+modeldir
        else:
            self.modeldir = modeldir
        self.modelname = model
        self.user = user
        self.debug = debug
        if not os.path.isdir(os.path.join(self.modeldir,self.modelname)):
            raise ValueError("no model dir for '%s'"%(self.modelname))
        try:
            mod,mod_shelf_fd = shelf_open(os.path.join(self.modeldir,self.modelname,".model"), "r")
        except:
            raise ValueError("no model db for '%s'"%(self.modelname))
        self.moddb = {}
        for k in mod:
            self.moddb[k] = mod[k]
        if (self.user == self.moddb['owner']):
            self.usertype = 'owner'
        else:
            self.usertype = 'app'

        shelf_close(mod, mod_shelf_fd)
        resolvers.Access.__init__(self)
        ev_filters.Access.__init__(self)

    def ask(self,  
        context=[],
        view=None,
        resolver=None,
        showcontexts=None):
        """

        :param context: a list giving the path of context identifiers
        :param view: either:
            - an identifier of a view in the context specified
            - a list of component identifiers or full path lists
            - None indicating that the values of all components in the context be returned
        :param resolver: 
            - a string containing the name of a resolver or
            - resolver is a dictionary containing information about resolver(s) to be used and arguments
                - the "resolver" key gives the name of a resolver to use, if not present the default resolver is used
                - the "evidence_filter" key specifies an evidence filter
                - eg 'evidence_filter' =  "all" returns all evidence,
                    - "last10" returns last 10 evidence items,
                    - "last1" returns most recent evidence item,
                    - None returns no evidence
        :param showcontexts: if True, a tuple is returned containing 
            (list of component objects, 
            list of contexts in the current context,
            list of views in the current context,
            list of subscriptions in the current context)
        :return: a list of component objects
        :raise:
        """
        logging.debug('ask: %s, %s, %s', context, view, resolver)
        self.curcontext = self._getcontextdir(context)
        contextinfo = self.getcontext(context)
        if 'perms' not in contextinfo:
            perms = []
            print 'no perms!'
        else:
            perms = contextinfo['perms']
        if self.usertype != 'owner':
            if self.user not in perms:
                raise ValueError("No permission for: %s in %s/%s" % (self.user, context, view))
        if self.usertype == 'app':
            if not self.checkpermission(context=context, componentid=None, app=self.user, permname='ask', permval=True):
                raise ValueError("No permission for: %s in %s/%s" % (self.user, context, view))
            
        try:
            comps,comps_shelf_fd = shelf_open(self.curcontext+"/.components", "r")
        except:
            comps = None
        try:
            views,views_shelf_fd = shelf_open(self.curcontext+"/.views", "r")
        except:
            views = None
        try:
            subs,subs_shelf_fd = shelf_open(self.curcontext+"/.subscriptions", "r")
        except:
            subs = None
        try:
            contexts = os.listdir(self.curcontext)
            contexts = [x for x in contexts if os.path.isdir(os.path.join(self.curcontext,x))]
        except OSError as e:
            raise ValueError("Component/Context not found: %s/%s" % (context, view))
        cidlist = []
        cobjlist = []

        if isinstance(view, basestring):
            if views != None:
                if view not in views:
                    raise ValueError('"%s" view not found'%(view))
                cidlist = views[view].component_list
            else:
                raise ValueError('"%s" view not found'%(view))
        elif type(view) is ListType:
            cidlist = view
        elif view == None: 
            if comps != None:
                cidlist = comps
        else:
            raise TypeError('view "%s" has unknown type'%(repr(view)))
        self.theresolver = None
        resolver_args = {}
        if resolver != None: 
            if type(resolver) == type({}):
                resolver_args = resolver.copy()
                resolver = resolver_args.get('resolver') # extract the string name of the resolver function
            if resolver != None:
                if resolver in self.resolverlist:
                    self.theresolver = self.resolverlist[resolver]
                else:
                    raise ValueError('unknown resolver "%s"'%(repr(resolver)))
        for cid in cidlist:
            if isinstance(cid, basestring):
                if comps != None:
                    if cid in comps:
                        if 'resolver' in comps[cid].__dict__:
                            compresname = comps[cid].resolver
                        else:
                            compresname = None
                        if 'resolver' in contextinfo:
                            contresname = contextinfo['resolver']
                        else:
                            contresname = None
                        if self.theresolver != None:
                            compresolver = self.theresolver
                        elif compresname == None:
                            if contresname != None:
                                if contresname in self.resolverlist:
                                    compresolver = self.resolverlist[contresname]
                                else:
                                    raise ValueError('unknown resolver "%s"'%(contresname))
                            else:
                                compresolver = self.resolverlist["default"]
                        elif compresname in self.resolverlist:
                            compresolver = self.resolverlist[compresname]
                        else:
                            raise ValueError('unknown resolver "%s"'%(compresname))
                        cobjlist.append(compresolver(model=self, 
                                                component=comps[cid],
                                                context=context, 
                                                resolver_args=resolver_args)
                                        )
                    else:
                        raise ValueError('component "%s" not in view "%s" (%s)'%(cid,view,cidlist))
                else:
                    raise ValueError('component "%s" not found'%(cid))
            elif type(cid) is ListType:
                vcontext = self._getcontextdir(cid[:-1])
                last_cid = cid[-1]
                try:
                    vcomps,vcomps_shelf_fd = shelf_open(vcontext+"/.components", "r")
                except:
                    raise ValueError('context "%s" not in view "%s"'%(last_cid,repr(view)))
                if last_cid in vcomps:
                    if 'resolver' in vcomps[last_cid].__dict__ :
                        compresname = vcomps[last_cid].resolver
                    else:
                        compresname = None
                    if self.theresolver != None:
                        compresolver = self.theresolver
                    elif compresname == None:
                        compresolver = self.resolverlist["default"]
                    elif compresname in self.resolverlist:
                        compresolver = self.resolverlist[compresname]
                    else:
                        raise ValueError('unknown resolver "%s"'%(compresname))
                    cobjlist.append(compresolver(model=self, component=vcomps[last_cid], \
                                context=context, resolver_args=resolver_args))
                else:
                    raise ValueError('component "%s" not in view "%s"'%(last_cid,repr(view)))
                shelf_close(vcomps, vcomps_shelf_fd)
                    
        if comps != None:
            shelf_close(comps, comps_shelf_fd)
        if views != None:
            theviews = {}
            for v in views:
                theviews[v] = views[v]
            shelf_close(views, views_shelf_fd)
        else:
            theviews = None
        if subs != None:
            thesubs = {}
            for s in subs:
                thesubs[s] = subs[s]
            shelf_close(subs, subs_shelf_fd)
        else:
            thesubs = None
        if showcontexts:
            return (cobjlist, contexts, theviews, thesubs)
        return cobjlist

    def export_model(self,
        context=[],
        resolver=None,
        level=None):
        """Export a model to a json representation

        :param context: a list giving the path of context identifiers this is the root of the um tree to export
        :param resolver: 
            - a string containing the name of a resolver or
            - a dictionary containing information about resolver(s) to be used and arguments
            - the "resolver" key gives the name of a resolver to use, if not present the default resolver is used
            - the "evidence_filter" key specifies an evidence filter
            - eg 'evidence_filter' =   "all" returns all evidence,
                - "last10" returns last 10 evidence items,
                - "last1" returns most recent evidence item,
                - None returns no evidence
        :param level:

        :return: returns a JSON encoded representation of the um tree
        :raise:
        """
        self.curcontext = self._getcontextdir(context)
        contextinfo = self.getcontext(context)
        perms = contextinfo['perms']
        if self.usertype != 'owner':
            if self.user not in perms:
                raise ValueError("No permission for: %s in %s/%s" % (self.user, context, view))
        if self.usertype == 'app':
            if not self.checkpermission(context=context, componentid=None, app=self.user, permname='ask', permval=True):
                raise ValueError("No permission for: %s in %s/%s" % (self.user, context, view))

        self.theresolver = None
        resolver_args = {}
        if resolver != None: 
            if type(resolver) == type({}):
                resolver_args = resolver.copy()
                resolver = resolver_args.get('resolver') # extract the string name of the resolver function
            if resolver != None:
                if resolver in self.resolverlist:
                    self.theresolver = self.resolverlist[resolver]
                else:
                    raise ValueError('unknown resolver "%s"'%(repr(resolver)))

        try:
            comps,comps_shelf_fd = shelf_open(self.curcontext+"/.components", "r")
        except:
            comps = None
        try:
            views,views_shelf_fd = shelf_open(self.curcontext+"/.views", "r")
        except:
            views = None
        try:
            subs,subs_shelf_fd = shelf_open(self.curcontext+"/.subscriptions", "r")
        except:
            subs = None
        try:
            contexts = os.listdir(self.curcontext)
            contexts = [x for x in contexts if os.path.isdir(self.curcontext+"/"+x)]
        except OSError as e:
            raise ValueError("Context not found: %s" % (context))
        
        if contexts != []:
            thecontexts = {}
            for con in contexts:
                thecontexts[con] = self.export_model(context=context+[con], resolver=resolver_args, level= level if level != None else 1)
        else:
            thecontexts = None

        if self.theresolver != None:
            compresolver = self.theresolver
        else:
            compresolver = self.resolverlist["default"]
        if comps != None:
            thecomps = {}
            for c in comps:
                #evlist = comps[c].filterevidence(model=self, context=context, resolver=resolver)
                #evlist2 = []
                #if type(evlist) == type([]):
                #   for e in evlist:
                #       evlist2.append(e.__dict__)
                #comps[c].__dict__["evidencelist"] = evlist2
                #thecomps[c] = comps[c].__dict__
                #thecomps[c]["evidencelist"] = evlist2
                thecomps[c] = compresolver(model=self, context=context, component=comps[c], resolver_args=resolver_args).__dict__
            shelf_close(comps, comps_shelf_fd)
        else:
            thecomps = None
        if subs != None:
            thesubs = {}
            for s in subs:
                thesubs[s] = subs[s]
            shelf_close(subs, subs_shelf_fd)
        else:
            thesubs = None
        if views != None:
            theviews = {}
            for v in views:
                theviews[v] = views[v].__dict__
            shelf_close(views, views)
        else:
            theviews = None

        contextdict = {"contextinfo":contextinfo, "contexts":thecontexts, "components":thecomps, "views":theviews, "subs":thesubs}
        if level == None:
            return json.dumps(contextdict)
        else:
            return contextdict

    def import_model(self, 
        context=[], 
        partial_model=None):
        """Import a model

        :param context: context to import partial model to if None, use root of model
        :param partial_model: 
            - string containing JSON representation of model dictionary OR
            - a dictionary with elements:
                - contextinfo - Description, Identifier, perms, resolver
                - contexts - sub contexts
                - components
                - views
                - subs
        
        :return:
        
        :raise:
        """
        self.curcontext = self._getcontextdir(context)
        contextinfo = self.getcontext(context)
        perms = contextinfo['perms']
        if self.usertype != 'owner':
            if not (self.user in perms):
                raise ValueError("No permission for: %s in %s/%s" % (self.user, context, view))
        if self.usertype == 'app':
            if not self.checkpermission(context=context, componentid=None, app=self.user, permname='tell', permval=True):
                raise ValueError("No permission for: %s in %s/%s" % (self.user, context, view))
            
        if partial_model == None:
            return
        if isinstance(partial_model, basestring):
            newmodel = json.loads(partial_model)
        else:
            newmodel = partial_model
        cinfo = newmodel['contextinfo']
        newcontext = context+[cinfo['Identifier']]
        if not self.existcontext(newcontext):
            con = Context(Identifier=cinfo['Identifier'], Description=cinfo['Description'], resolver=cinfo['resolver'], perms=cinfo['perms'])
            self.mkcontext(context=context, contextobj=con)
        else:
            print newcontext, "exists"
        for compname, comp in newmodel['components'].items():
            newcobj = Component()
            for k,v in comp.items():
                #print "comp items:", k, v
                if k not in ["evidencelist","value"]:
                    newcobj.__dict__[k] = v
            self.mkcomponent(newcontext, newcobj)
            comp["evidencelist"].reverse()
            for ev in comp["evidencelist"]:
                evv = Evidence(evidence_type="explicit") # evidence type will be overwritten by imported evidence
                for k,v in ev.items():
                    evv.__dict__[k] = v
                print "=>> tell", newcontext, compname, evv.__dict__
                self.tell(context=newcontext, componentid=compname, evidence=evv, dosubs=False)
        for viewname, view in newmodel['views'].items():
            newview = View()
            for k,v in view.items():
                newview.__dict__[k] = v
            try:
                self.mkview(newcontext, newview)
            except:
                print "View exists:", newcontext, newview
            print ">>VIEW", viewname, view
        for subname, sub in newmodel['subs'].items():
            for k,v in sub.items():
                print ">>SUB", subname, k, v
                self.subscribe(context=newcontext, view=[subname], subscription=v)
        if newmodel['contexts'] != None:
            for contextname, cont in newmodel['contexts'].items():
                print ">>CONTEXT", contextname, cont
                self.import_model(newcontext, cont)
        
        return newmodel
    
    def tell(self, 
        context=[],
        componentid=None,
        evidence=None,   # evidence obj
        dosubs=True):
        """Tell the model something

        :param context: a list giving the path to the required context
        :param componentid: identifier of the component
        :param evidence: (:class:`Evidence`) object to add to the component
        :return:
        :raise:
        """
        logging.debug('context: %s, componentid %s, evidence %s, dosubs %s', repr(context), repr(componentid), repr(evidence), dosubs)
        if componentid == None:
            raise ValueError("tell: component id is None")
        if evidence == None:
            raise ValueError("tell: no evidence provided")
        if not isinstance(componentid, basestring):
            raise ValueError("tell: component id is not string type")
        self.curcontext = self._getcontextdir(context)
        contextinfo = self.getcontext(context)
        logging.debug('contextinfo %s',contextinfo)
        perms = contextinfo['perms']
        if self.usertype != 'owner':
            if not (self.user in perms):
                raise ValueError("No permission for: %s in %s" % (self.user, context))
        if self.usertype == 'app':
            if not self.checkpermission(context=context, componentid=None, app=self.user, permname='tell', permval=True):
                raise ValueError("No permission for: %s in %s/%s" % (self.user, context, componentid))
        comps,comps_shelf_fd = shelf_open(self.curcontext+"/.components", "w")

        if componentid in comps:
            evidence.creation_time = time.time()
            evidence.owner = self.user
            cobj = comps[componentid]
            if cobj.value_type == "enum":
                if evidence.value not in cobj.value_list:
                    raise ValueError("tell: value '%s' not in value list %s for component '%s' of type 'enum'" % (evidence.value, repr(cobj.value_list), componentid))
            try:
                evdb,evdb_fd = shelf_open(self.curcontext+"/.evidence", "w")
            except:
                raise ValueError("tell: no evidence db for %s"%(self.curcontext))
            if componentid not in evdb:
                raise ValueError("tell: no evidence db entry for %s"%(componentid))
            evcount = evdb[componentid]+1
            evdb[componentid] = evcount
            evidence.source = self.user
            evdb["%s:%d"%(componentid,evcount)] = evidence
            shelf_close(evdb, evdb_fd)
            if dosubs:
                self.checksubs(context, componentid)
        else:
            raise ValueError("tell: component id %s not found"%(componentid))
        shelf_close(comps, comps_shelf_fd)
        return None

    def set_goals(self, 
        context=[],
        componentid=None,
        goals=None):
        """set the goal list for a component. requires "tell" permission

        :param context: a list giving the path to the required context
        :param componentid: identifier of the component
        :param goals: list of goal component paths
        :return:
        :raise:
        """
        if componentid == None:
            raise ValueError("set_goals: component id is None")
        if goals == None:
            raise ValueError("set_goals: no goal list provided")
        self.curcontext = self._getcontextdir(context)
        contextinfo = self.getcontext(context)
        perms = contextinfo['perms']
        if self.usertype != 'owner':
            if self.user not in perms:
                raise ValueError("No permission for: %s in %s" % (self.user, context))
        if self.usertype == 'app':
            if not self.checkpermission(context=context, componentid=None, app=self.user, permname='tell', permval=True):
                raise ValueError("No permission for: %s in %s/%s" % (self.user, context, componentid))
        comps,comps_shelf_fd = shelf_open(self.curcontext+"/.components", "w")
        if componentid in comps:
            cobj = comps[componentid]
        else:
            raise ValueError("set_goals: component id %s not found"%(componentid))
        cobj.goals = goals
        comps[componentid] = cobj
        shelf_close(comps, comps_shelf_fd)
        return None

    def checksubs(self, context, componentid):
        """check subscriptions for componentid in current context. implemented in active

        :param context: a list giving the path to the required context
        :param componentid: identifier of the component
        :return:
        :raise:
        """
        pass
    
    def getcontext(self, context=[], getsize=False):
        """
        :param context: a list giving the path to the required context
        :return: a dictionary containing information associated with the context
        :raise:
        """
        def getcontextsize(dir="."):
            size = 0
            dir += "/"
            for f in glob.glob(dir+".*") + glob.glob(dir+"*"):
                if os.path.isdir(f):
                    size += getcontextsize(f)
                else:
                    size += os.stat(f).st_size
            return size

        try:
            contextdb,contextdb_shelf_fd = shelf_open(self._getcontextdir(context)+"/.context", "r")
        except:
            return {} # root context - FIX
        contextinfo = {}
        for k in contextdb:
            contextinfo[k] = contextdb[k]
        shelf_close(contextdb, contextdb_shelf_fd)
        if getsize:
            contextinfo['size'] = getcontextsize(self._getcontextdir(context))

        logging.debug('Context is %s', contextinfo)
        return contextinfo

    def setcontext(self, context=[], **kwargs):
        """Set the context

        :param context: a list giving the path to the required context 
        :param \*\*kwargs: a set of keyword arguments giving information to associate with the context

        :return:
            True on success
            False if the context doesn't exist
        """
        try:
            contextdb, contextdb_shelf_fd = shelf_open(self._getcontextdir(context)+"/.context", "w")
        except:
            return False
        for k,v in kwargs.items():
            contextdb[k] = v
        shelf_close(contextdb, contextdb_shelf_fd)
        return True


    def existcontext(self, context=[]):
        """
        :param context: a list giving the path to the required context 
    
        :return:
            True if context exists
            False if the context doesn't exist
        """
        return os.path.exists(self._getcontextdir(context)+"/.context")

    def mkcomponent(self, 
        context=[],
        componentobj=None):
        """Make a new component in a given context

        :param context: a list giving the path to the required context 
        :param componentobj: (:class:`Component`) a Component object with the context
        :return:
            None on success
            a string error message on error
        """
        if componentobj == None:
            return "component object is None"
        if componentobj.value != None:
            return "component value specified as %s at creation - not allowed"%(componentobj.value)
        self.curcontext = self._getcontextdir(context)
        comps, comps_shelf_fd = shelf_open(self.curcontext+"/.components", "w")
        if componentobj.Identifier in comps:
            return "component %s already exists"%(componentobj.Identifier)
        comps[componentobj.Identifier] = componentobj
        shelf_close(comps, comps_shelf_fd)
        try:
            evdb,evdb_fd = shelf_open(self.curcontext+"/.evidence", "w")
        except:
            raise ValueError("mkcomponent: no evidence db for %s"%(self.curcontext))
        if componentobj.Identifier in evdb:
            raise ValueError("mkcomponent: evidence db entry for %s already present"%(componentid))
        evdb[componentobj.Identifier] = 0
        shelf_close(evdb, evdb_fd)
        return None # all ok

    def delcomponent(self,
        context= [],
        componentid=None):
        """Delete an existing component in a given context
        
        :param context: a list giving the path to the required context 
        :param id: the id for a componen
        
        :return:
            None on success
            a string error message on error
        """
        if componentid == None:
            return "no component specified for deletion"
        self.curcontext = self._getcontextdir(context)
        comps, comps_shelf_fd = shelf_open(self.curcontext+"/.components", "w")
        if componentid in comps:
            del comps[componentid]
        else:
            return "no component matched $s"%(componentid)

        shelf_close(comps, comps_shelf_fd)
        try:
            evdb,evdb_fd = shelf_open(self.curcontext+"/.evidence", "w")
        except:
            raise ValueError("delcomponent: no evidence db for %s"%(self.curcontext))
        if componentid not in evdb:
            raise ValueError("mkcomponent: no evidence db entry for %s "%(componentid))
        evcount = evdb[componentid]
        for evcount in range(evdb[componentid]):
            del evdb["%s:%d"%(componentid, evcount+1)]
        del evdb[componentid]
        shelf_close(evdb, evdb_fd)
        return None # all ok

    def registerapp(self, app=None, desc="", password=None):
        """registers a password for an app
        
        :param app: app name is a string (needs checking TODO)
        :param desc: the app description string
            app passwords are stored at the top level .model db
        
        :return: returns a dictionary containing description and password(access key)
        """
        if self.usertype != 'owner':
            raise ValueError("registerapp: must be owner to set password for %s" % (app))
        p = hashlib.md5()
        if password == None:
            password = ''
        p.update(password)
        if app not in self.moddb['apps']:
            self.moddb['apps'][app] = {}
        self.moddb['apps'][app]['description'] = desc
        self.moddb['apps'][app]['password'] = p.hexdigest()

        mod, mod_shelf_fd = shelf_open(os.path.join(self.modeldir,self.modelname,".model"), "w")
        mod['apps'] = self.moddb['apps']
        shelf_close(mod, mod_shelf_fd)
        return self.moddb['apps'][app]

    def deleteapp(self, app=None):
        """deletes an app
        """
        if app not in self.moddb['apps']:
            raise ValueError("deleteapp: app %s not registered"%(app))
        if self.usertype != 'owner':
            raise ValueError("deleteapp: must be owner to delete app %s" % (app))
        del self.moddb['apps'][app]
        mod, mod_shelf_fd = shelf_open(os.path.join(self.modeldir,self.modelname,".model"), "w")
        mod['apps'] = self.moddb['apps']
        shelf_close(mod, mod_shelf_fd)

    def listapps(self):
        """returns an dictionary of apps that are registered
            key is app name, 'description' is app description
        """
        apps = {}

        for app in self.moddb['apps']:
            apps[app] = {}
            apps[app]['description'] = self.moddb['apps'][app]['description']
        return apps

    def setpermission(self, context=None, componentid=None, app=None, permissions={}):
        """sets ask/tell permission for a context (if componentid is None) or
                a component
        """
        if self.usertype != "owner":
            raise ValueError("must be model owner to set permissions")
        if componentid == None:
            try:
                contextdb,contextdb_shelf_fd = shelf_open(self._getcontextdir(context)+"/.context", "w")
            except:
                raise ValueError("setpermission: no context db for %s"%(context))
            perms = contextdb['perms']
            if app not in self.moddb['apps']:
                raise ValueError("setpermission: app %s not registered"%(app))
            if app not in perms:
                perms[app] = {}
            for k,v in permissions.items():
                perms[app][k] = v
            contextdb['perms'] = perms
            shelf_close(contextdb, contextdb_shelf_fd)
            

    def getpermission(self, context=None, componentid=None, app=None):
        """gets permissions for a context (if componentid is None) or a component # not implemented yet ####
        
        :return: a tuple (ask,tell)
        """
        if app not in self.moddb['apps']:
            raise ValueError("getpermission: app %s not registered"%(app))
        if componentid == None:
            try:
                contextdb,contextdb_shelf_fd = shelf_open(self._getcontextdir(context)+"/.context", "r")
            except:
                raise ValueError("getpermission: no context db for %s"%(context))
            perms = contextdb['perms']
            shelf_close(contextdb, contextdb_shelf_fd)
            return perms.get(app)

    def checkpermission(self, context=None, componentid=None, app=None, permname=None, permval=None):
        """gets permissions for a context (if componentid is None) or a component # not implemented ####
        :return:
            True is the app has the nominated permission
            False if not
        """
        perms = self.getpermission(context=context, componentid=componentid, app=app)
        if permname in perms:
            return permval == perms[permname]
        return False

    def setcomponentoption(self, context, componentid, **kwargs):
        """

        :param context: a list giving the path to the required context 
        :param componentid: name of the component to set items on
        :param \*\*kwargs: a set of keyword arguments giving information to associate with the component
    
        :return:
            True on success
            False if the component doesn't exist
        """
        ok = False
        # try:
        if True:
            comps, comps_shelf_fd = shelf_open(self._getcontextdir(context)+"/.components", "w")
            if componentid in comps:
                comp = comps[componentid]
                for k,v in kwargs.items():
                    comp.__dict__[k] = v
                comps[componentid] = comp
            shelf_close(comps, comps_shelf_fd)
            ok = True
        try:
            pass
        except:
            
            try:
                shelf_close(comps, comps_shelf_fd)
            except:
                pass
        return ok

    def setresolver(self, context, componentid, resolver):
        return self.setcomponentoption(context,componentid,resolver=resolver)

    def getresolvers(self):
        return list(self.resolverlist.keys())

    def mkcontext(self, 
        context= [],
        contextobj=None):
        """Make a new context in a given context

        :param context: a list giving the path to the context that will hold the new context 
        :param contextobj: a Context object
        
        :return: True if created ok, False otherwise
        """
        self.curcontext = self._getcontextdir(context)
        self.curcontext += "/"+contextobj.Identifier
        try:
            os.mkdir(self.curcontext)
        except:
            return False
        con, con_shelf_fd = shelf_open(self.curcontext+"/.context", "c")
        for k,v in contextobj.__dict__.items():
            con[k] = v
        shelf_close(con, con_shelf_fd)
        comps, comps_shelf_fd = shelf_open(self.curcontext+"/.components", "c")
        shelf_close(comps, comps_shelf_fd)
        evdb,evdb_fd = shelf_open(self.curcontext+"/.evidence", "c")
        shelf_close(evdb, evdb_fd)
        views, views_shelf_fd = shelf_open(self.curcontext+"/.views", "c")
        shelf_close(views, views_shelf_fd)
        subs, subs_shelf_fd = shelf_open(self.curcontext+"/.subscriptions", "c")
        shelf_close(subs, subs_shelf_fd)
        return True

    def delcontext(self,
        context=[]):
        """Delete a context saves an archive of the context (debug)
        """
        okchars = string.ascii_letters+string.digits+'_'
        for c in context:
            ok = False not in [x in okchars for x in list(c)]
            if not ok:
                raise ValueError('bad character in context name')
        ctxtdir = [self.modelname]+context
        delcontext = ctxtdir[-1]  # directory to remove
        
        try:
            ctxtbase = [self.modeldir] + ctxtdir
            ctxtbase = ctxtbase[:-1]
            #print ">>> chdir ", os.path.join(*ctxtbase)
            os.chdir(os.path.join(*ctxtbase))
        except:
            return False
        
        import tarfile, time
        tarname = "%s-%s.tgz" % (time.strftime("%Y%m%d%H%M", time.localtime()), delcontext)
        #print ">>> ", tarname
        tar = tarfile.open(os.path.join(*[self.modeldir, tarname]), "w:gz")
        tar.add(delcontext)
        tar.close()

        #print '>>> os.system("rm -rf ' + os.path.join(*([self.modeldir]+ctxtdir))+'"'
        os.system("rm -rf "+os.path.join(*([self.modeldir]+ctxtdir)))
        return True

    def mkview(self, 
        context= [],
        viewobj=None):
        """Make a new view in a given context

        :param context: a list giving the path to the required context 
        :param viewobj: a View object
        """
        if viewobj == None:
            raise ValueError('view object is None')
        view_id = viewobj.Identifier
        self.curcontext = self._getcontextdir(context)
        views, views_shelf_fd = shelf_open(self.curcontext+"/.views", "w")
        if view_id in views:
            raise ValueError("view %s already exists"%(viewobj.Identifier))
        views[view_id] = viewobj
        shelf_close(views, views_shelf_fd)
        return None

    def delview(self,
        context=[],
        viewid=None):
        """Delete an existing view within a given context

        :param context: a list giving the path to the required context 
        :param viewid: view identifier
    
        :return:
            on success, None
            on failure, a string reporting the problem
        """
        if viewid == None:
            return "no view specified for deletion"
        self.curcontext = self._getcontextdir(context)
        views, views_shelf_fd = shelf_open(self.curcontext+"/.views", "w")
        if viewid in views:
            del views[viewid]
        else:
            return "no view matched $s"%(viewid)
        shelf_close(views, views_shelf_fd)
        return None # all ok

    def _getcontextdir(self, context):
        if context == None:
            context = []
        if isinstance(context, basestring):
            context = [x for x in context.split(os.sep) if x not in ['','.','..']]
        if not (type(context) is ListType):
            raise ValueError("Context <%s> is wrong type"%(context))
        ctxt = [self.modeldir, self.modelname]+context
        return os.path.join(*ctxt)




import os, sys

from oauth2client.file import Storage
from oauth2client.client import Storage, Credentials, OAuth2WebServerFlow, AccessTokenRefreshError
from oauth2client.tools import run

import httplib
import httplib2
import webapp2
from gaesessions import get_current_session

from personis import client
import yaml
import time
import logging
import pickle
from collections import OrderedDict


# items must be listed in correct order, so use OrderedDict and add manually.
item_list = OrderedDict()

item_list['fruit'] = {'category': 'food', 'icon':'/image/fruit.jpg'}
item_list['running'] = {'category': 'activity', 'icon': '/image/running.jpg'}
item_list['vegetables'] = {'category': 'food', 'icon': '/image/vegetables.jpg'}
item_list['walking'] = {'category': 'activity', 'icon': '/image/walking.jpg'}

        # running - http://www.flickr.com/photos/good_day/20723337/in/photostream/
        # walking - http://www.flickr.com/photos/o5com/5081595200/in/photostream/
        # fruit - http://www.flickr.com/photos/doug88888/2780642603/in/photostream/
        # vegetables - http://www.flickr.com/photos/suckamc/2488644619/in/photostream/


class LogLlum(webapp2.RequestHandler):


    def install_contexts(self, um):
        try:
            reslist = um.ask(context=['Apps','Logging'], view=['logged_items'])
            return
        except:
            pass

        context = ['Apps']
        ctx_obj = client.Context(Identifier="Logging",
                  Description="The logging app",
                  perms={'ask':True, 'tell':True,
                  "resolvers": ["all","last10","last1","goal"]},
                  resolver=None, objectType="Context")
        print "Creating logging context "
        try:
            um.mkcontext(context,ctx_obj)
        except:
            logging.debug('already have logging context')
        context.append('Logging')

        value_list = [i for i in item_list.keys()] + [i+'-' for i in item_list.keys()]
        cobj = client.Component(Identifier="logged_items", component_type="activity", value_type="enum", 
                                       value_list=value_list, resolver=None ,Description="All the items logged")
        try:
            um.mkcomponent(context=context, componentobj=cobj)
        except:
            logging.debug('already have logged_items')

    def do_login(self):
        session = get_current_session()
        oauthconf = self.app.config.get('oauthconf')
        flow = OAuth2WebServerFlow(client_id=oauthconf['client_id'],
                        client_secret=oauthconf['client_secret'],
                        scope='https://www.personis.info/auth/model',
                        user_agent='Log-llum/1.0',
                        auth_uri=oauthconf['personis_uri']+'/authorize',
                        token_uri=oauthconf['personis_uri']+'/request_token')
        callback = oauthconf['callback']
        authorize_url = flow.step1_get_authorize_url(callback)
        session['flow'] = pickle.dumps(flow)
        return self.redirect(authorize_url)


    def authorized(self):
        session = get_current_session()
        if not 'code' in self.request.params:
            self.abort(400, detail='no code param')

        code = self.request.params['code']
        flow = session.get('flow')
        flow = pickle.loads(flow)
        if not flow:
            raise IOError()
        session['flow'] = None
        credentials = flow.step2_exchange(self.request.params, http = httplib2.Http(disable_ssl_certificate_validation=True))
        oauthconf = self.app.config.get('oauthconf')
        c = client.Connection(uri = oauthconf['personis_uri'], credentials = credentials)
        session['connection'] = pickle.dumps(c)
        um = client.Access(connection=c, http = httplib2.Http(disable_ssl_certificate_validation=True))
        self.install_contexts(um)
        self.redirect('/')

    def log_me(self):
        if self.request.method != 'POST':
            self.abort(400, detail="POST only")
        session = get_current_session()
        if session.get('connection') == None:
            self.abort(400, detail='Log in first.')
        logging.debug('logme %s', self.request.get('item'))
        connection = pickle.loads(session.get('connection'))
        um = client.Access(connection=connection, http = httplib2.Http(disable_ssl_certificate_validation=True), test=False)
        item = self.request.get('item')
        ev = client.Evidence(source='llum-log', evidence_type="explicit", value=item, time=time.time())
        um.tell(context=['Apps','Logging'], componentid='logged_items', evidence=ev)
        self.response.write(item)

    def get(self):
        session = get_current_session()
        if session.get('connection') == None:
            logging.debug('no connection found. logging in')
            return self.redirect('/do_login')
        connection = pickle.loads(session.get('connection'))
        um = client.Access(connection=connection, test=False, http = httplib2.Http(disable_ssl_certificate_validation=True))
        self.install_contexts(um)
        try:
            reslist = um.ask(context=["Personal"],view=['firstname', 'picture'])
        except AccessTokenRefreshError as e:
            logging.info('access token refresh error %s', e)
            return self.redirect('/do_login')

        ret = '''
    <!DOCTYPE html>
<html>
    <head>
        <meta charset="utf-8" />
        <meta name="viewport" content="width=device-width, initial-scale=1" />
        <link rel="stylesheet" href="https://ajax.aspnetcdn.com/ajax/jquery.mobile/1.1.1/jquery.mobile-1.1.1.min.css" />
        <script src="https://ajax.googleapis.com/ajax/libs/jquery/1.8.0/jquery.min.js"></script>
        <script src="https://ajax.aspnetcdn.com/ajax/jquery.mobile/1.1.1/jquery.mobile-1.1.1.min.js"></script>
        <link rel="stylesheet" href="/static/my.css"/>
        <script src="/static/my.js"></script>
    </head>
    <body>
        <!-- Home -->
        <div data-role="page" id="page1">
            <div data-theme="a" data-role="header" data-position="fixed">
            <h3>My Health Logger</h3>
            <a href=""><img src='{0[user_icon]}' style="height: 50px"/></a>
            </div>
            <div data-role="content" style="padding: 15px">
                <div class="ui-grid-a">
                    <div class="ui-block-a" align="center"><h2>Food</h2></div>
                    <div class="ui-block-b" align="center"><h2>Activity</h2></div>'''.format({'user_icon':reslist[1].value })
        l = 'a'
        for k, v in item_list.items():
            ret = ret + ''' <div class="ui-block-{0[l]}">
                <div class="imageC">
                    <img class="wrapper"  src='{0[pic]}' alt='{0[name]}'/>
                </div>
            </div>'''.format({'l': l,'name': k.strip(), 'pic': v['icon']})
            l = 'a' if l == 'b' else 'b'
        ret = ret + ''' 
                </div>
                <button id="undo">Undo</button>
            <div data-role="collapsible" data-collapsed="false" id="coll" data-theme='c' data-content-theme="c">
              <h3>Logged items</h3>
                <p class='loggedItem'></p>
            </div>
        </div>
        </div>
    </body>
</html>'''
        self.response.write(ret)


httplib2.debuglevel=5
logging.basicConfig(level=logging.DEBUG)
config = {}
config['oauthconf'] = yaml.load(file('oauth.ae.yaml','r'))

app = webapp2.WSGIApplication([
        webapp2.Route(r'/do_login', handler=LogLlum, handler_method='do_login'),
        webapp2.Route(r'/authorized', handler=LogLlum, handler_method='authorized'),
        webapp2.Route(r'/log_me', handler=LogLlum, handler_method='log_me'),
        (r'/', LogLlum)
    ],
    debug=False, config=config)






import os, sys

import httplib, oauth2
from optparse import OptionParser
import httplib2

from oauth2client.file import Storage
from oauth2client.client import Storage, Credentials, OAuth2WebServerFlow
from oauth2client.tools import run

import cherrypy
import wsgiref.handlers
import ConfigParser

from personis import client
import yaml
import time

item_list = {'apple':{'icon':'http://appleadayproject.files.wordpress.com/2011/03/apple-full2.jpg'},
            'pear':{'icon': 'http://4.bp.blogspot.com/-IgzE0L2YSdg/T1Pg-z8t6-I/AAAAAAAAAhU/cWfds0ulbLI/s1600/Pear.jpg'}, 
            'banana':{'icon': 'http://upload.wikimedia.org/wikipedia/commons/thumb/8/8a/Banana-Single.jpg/220px-Banana-Single.jpg'}, 
            'orange':{'icon': 'http://freeimagesarchive.com/data/media/38/7_orange.jpg'},
            'kiwi':{'icon': 'http://1.bp.blogspot.com/-XK8RbZ1MFz8/T88vJy9THfI/AAAAAAAABts/FWCuZptW_d0/s1600/kiwi+fruit.jpg'},
            'grape':{'icon': 'http://upload.wikimedia.org/wikipedia/commons/thumb/b/bb/Table_grapes_on_white.jpg/220px-Table_grapes_on_white.jpg'}}

class LogLlum(object):

    def __init__(self, oauthconf):
        self.oauthconf = yaml.load(file(oauthconf,'r'))

    @cherrypy.expose
    def do_login(self):
        flow = OAuth2WebServerFlow(client_id=self.oauthconf['client_id'],
                        client_secret=self.oauthconf['client_secret'],
                        scope='https://www.personis.info/auth/model',
                        user_agent='Log-llum/1.0',
                        auth_uri=self.oauthconf['personis_uri']+'/authorize',
                        token_uri=self.oauthconf['personis_uri']+'/request_token')
        callback = self.oauthconf['callback']
        authorize_url = flow.step1_get_authorize_url(callback)
        cherrypy.session['flow'] = flow
        raise cherrypy.HTTPRedirect(authorize_url)

    def install_contexts(self, um):
        try:
            reslist = um.ask(context=['Apps','Logging'])
            return
        except:
            pass

        context = ['Apps']
        ctx_obj = client.Context(Identifier="Logging",
                  Description="The logging app",
                  perms={'ask':True, 'tell':True,
                  "resolvers": ["all","last10","last1","goal"]},
                  resolver=None, objectType="Context")
        print "Creating logging context "
        um.mkcontext(context,ctx_obj)
        context.append('Logging')

        cobj = client.Component(Identifier="logged_items", component_type="activity", value_type="enum", 
                                       value_list=[i for i in item_list.keys()], resolver=None ,Description="All the items logged")
        um.mkcomponent(context=context, componentobj=cobj)

    @cherrypy.expose
    def authorized(self, code, state=None):
        flow = cherrypy.session.get('flow')
        if not flow:
            raise IOError()
        p = httplib2.ProxyInfo(proxy_type=httplib2.socks.PROXY_TYPE_HTTP_NO_TUNNEL, proxy_host='www-cache.it.usyd.edu.au', proxy_port=8000)
        h = httplib2.Http(proxy_info=p)
        credentials = flow.step2_exchange(cherrypy.request.params, h)
        ht = httplib2.Http(proxy_info=p)
        um = client.Access(uri = self.oauthconf['personis_uri'], credentials = credentials, http = ht)
        self.install_contexts(um)
        cherrypy.session['um'] = um
        raise cherrypy.HTTPRedirect('/')

    @cherrypy.expose
    def log_me(self, item):
        if cherrypy.session.get('um') == None:
            raise cherrypy.HTTPError(400, 'Log in first.')
        um = cherrypy.session.get('um')
        ev = client.Evidence(source='llum-log', evidence_type="explicit", value=item, time=time.time())
        um.tell(context=['Apps','Logging'], componentid='logged_items', evidence=ev)
        raise cherrypy.HTTPRedirect('/')

    @cherrypy.expose
    def index(self):
        if cherrypy.session.get('um') == None:
            raise cherrypy.HTTPRedirect('/do_login')
        um = cherrypy.session.get('um')
        reslist = um.ask(context=["Personal"],view=['firstname', 'picture'])

        ret = '''<!DOCTYPE html>
<html>
    <head>
        <meta charset="utf-8" />
        <meta name="viewport" content="width=device-width, initial-scale=1" />
        <title>Logger</title>
        <link rel="shortcut icon" href="/favicon.ico" />
        <link rel="stylesheet" href="http://code.jquery.com/mobile/1.1.0/jquery.mobile-1.1.0.min.css" />
        <script src="http://ajax.googleapis.com/ajax/libs/jquery/1.7.1/jquery.min.js">
        </script>
        <script src="http://code.jquery.com/mobile/1.1.0/jquery.mobile-1.1.0.min.js">
        </script>
    </head>
    <body id="home-index-action" class="ot">
        <div id="homePage" data-role="page" data-theme="o" class="homeBody">
        <div class="mainContentPanel" data-role="content">
        <div id="OTHomeLogo">Hi {0[firstname]}, <img src="{0[user_icon]}" style='max-width:50px; max-height:50px' border="0" class="homeLogo">. Log something!</div>
        <div class="ui-grid-b">
        '''.format({'firstname': reslist[0].value, 'user_icon':reslist[1].value })
        for k, v in item_list.items():
            ret = ret + '''<div class="ui-block-a">
                <a class="wrapper" href="/log_me?item={0[name]}" id="ByName"><img style='max-width:100px; max-height:100px' src='{0[pic]}'/></a>
            </div>
            '''.format({'name': k, 'pic': v['icon']})
        ret = ret + '</div></div></div></body></html>'
        return ret

if __name__ == '__main__':
    httplib2.debuglevel=0
    parser = OptionParser()
    parser.add_option("-o", "--oauthconf",
          dest="oauthconf", metavar='FILE',
          help="Oauth Config file", default='oauth.yaml')
    (options, args) = parser.parse_args()
    cherrypy.quickstart(LogLlum(options.oauthconf),'/',config='global.conf')
    cherrypy.engine.start()
    cherrypy.engine.block()




# Copyright 2011 Google Inc. All Rights Reserved.

"""Multi-credential file store with lock support.

This module implements a JSON credential store where multiple
credentials can be stored in one file.  That file supports locking
both in a single process and across processes.

The credential themselves are keyed off of:
* client_id
* user_agent
* scope

The format of the stored data is like so:
{
  'file_version': 1,
  'data': [
    {
      'key': {
        'clientId': '<client id>',
        'userAgent': '<user agent>',
        'scope': '<scope>'
      },
      'credential': {
        # JSON serialized Credentials.
      }
    }
  ]
}
"""

__author__ = 'jbeda@google.com (Joe Beda)'

import base64
import errno
import logging
import os
import threading

from anyjson import simplejson
from client import Storage as BaseStorage
from client import Credentials
from locked_file import LockedFile

logger = logging.getLogger(__name__)

# A dict from 'filename'->_MultiStore instances
_multistores = {}
_multistores_lock = threading.Lock()


class Error(Exception):
  """Base error for this module."""
  pass


class NewerCredentialStoreError(Error):
  """The credential store is a newer version that supported."""
  pass


def get_credential_storage(filename, client_id, user_agent, scope,
                           warn_on_readonly=True):
  """Get a Storage instance for a credential.

  Args:
    filename: The JSON file storing a set of credentials
    client_id: The client_id for the credential
    user_agent: The user agent for the credential
    scope: string or list of strings, Scope(s) being requested
    warn_on_readonly: if True, log a warning if the store is readonly

  Returns:
    An object derived from client.Storage for getting/setting the
    credential.
  """
  filename = os.path.realpath(os.path.expanduser(filename))
  _multistores_lock.acquire()
  try:
    multistore = _multistores.setdefault(
        filename, _MultiStore(filename, warn_on_readonly))
  finally:
    _multistores_lock.release()
  if type(scope) is list:
    scope = ' '.join(scope)
  return multistore._get_storage(client_id, user_agent, scope)


class _MultiStore(object):
  """A file backed store for multiple credentials."""

  def __init__(self, filename, warn_on_readonly=True):
    """Initialize the class.

    This will create the file if necessary.
    """
    self._file = LockedFile(filename, 'r+b', 'rb')
    self._thread_lock = threading.Lock()
    self._read_only = False
    self._warn_on_readonly = warn_on_readonly

    self._create_file_if_needed()

    # Cache of deserialized store.  This is only valid after the
    # _MultiStore is locked or _refresh_data_cache is called.  This is
    # of the form of:
    #
    # (client_id, user_agent, scope) -> OAuth2Credential
    #
    # If this is None, then the store hasn't been read yet.
    self._data = None

  class _Storage(BaseStorage):
    """A Storage object that knows how to read/write a single credential."""

    def __init__(self, multistore, client_id, user_agent, scope):
      self._multistore = multistore
      self._client_id = client_id
      self._user_agent = user_agent
      self._scope = scope

    def acquire_lock(self):
      """Acquires any lock necessary to access this Storage.

      This lock is not reentrant.
      """
      self._multistore._lock()

    def release_lock(self):
      """Release the Storage lock.

      Trying to release a lock that isn't held will result in a
      RuntimeError.
      """
      self._multistore._unlock()

    def locked_get(self):
      """Retrieve credential.

      The Storage lock must be held when this is called.

      Returns:
        oauth2client.client.Credentials
      """
      credential = self._multistore._get_credential(
          self._client_id, self._user_agent, self._scope)
      if credential:
        credential.set_store(self)
      return credential

    def locked_put(self, credentials):
      """Write a credential.

      The Storage lock must be held when this is called.

      Args:
        credentials: Credentials, the credentials to store.
      """
      self._multistore._update_credential(credentials, self._scope)

    def locked_delete(self):
      """Delete a credential.

      The Storage lock must be held when this is called.

      Args:
        credentials: Credentials, the credentials to store.
      """
      self._multistore._delete_credential(self._client_id, self._user_agent,
          self._scope)

  def _create_file_if_needed(self):
    """Create an empty file if necessary.

    This method will not initialize the file. Instead it implements a
    simple version of "touch" to ensure the file has been created.
    """
    if not os.path.exists(self._file.filename()):
      old_umask = os.umask(0177)
      try:
        open(self._file.filename(), 'a+b').close()
      finally:
        os.umask(old_umask)

  def _lock(self):
    """Lock the entire multistore."""
    self._thread_lock.acquire()
    self._file.open_and_lock()
    if not self._file.is_locked():
      self._read_only = True
      if self._warn_on_readonly:
        logger.warn('The credentials file (%s) is not writable. Opening in '
                    'read-only mode. Any refreshed credentials will only be '
                    'valid for this run.' % self._file.filename())
    if os.path.getsize(self._file.filename()) == 0:
      logger.debug('Initializing empty multistore file')
      # The multistore is empty so write out an empty file.
      self._data = {}
      self._write()
    elif not self._read_only or self._data is None:
      # Only refresh the data if we are read/write or we haven't
      # cached the data yet.  If we are readonly, we assume is isn't
      # changing out from under us and that we only have to read it
      # once.  This prevents us from whacking any new access keys that
      # we have cached in memory but were unable to write out.
      self._refresh_data_cache()

  def _unlock(self):
    """Release the lock on the multistore."""
    self._file.unlock_and_close()
    self._thread_lock.release()

  def _locked_json_read(self):
    """Get the raw content of the multistore file.

    The multistore must be locked when this is called.

    Returns:
      The contents of the multistore decoded as JSON.
    """
    assert self._thread_lock.locked()
    self._file.file_handle().seek(0)
    return simplejson.load(self._file.file_handle())

  def _locked_json_write(self, data):
    """Write a JSON serializable data structure to the multistore.

    The multistore must be locked when this is called.

    Args:
      data: The data to be serialized and written.
    """
    assert self._thread_lock.locked()
    if self._read_only:
      return
    self._file.file_handle().seek(0)
    simplejson.dump(data, self._file.file_handle(), sort_keys=True, indent=2)
    self._file.file_handle().truncate()

  def _refresh_data_cache(self):
    """Refresh the contents of the multistore.

    The multistore must be locked when this is called.

    Raises:
      NewerCredentialStoreError: Raised when a newer client has written the
        store.
    """
    self._data = {}
    try:
      raw_data = self._locked_json_read()
    except Exception:
      logger.warn('Credential data store could not be loaded. '
                  'Will ignore and overwrite.')
      return

    version = 0
    try:
      version = raw_data['file_version']
    except Exception:
      logger.warn('Missing version for credential data store. It may be '
                  'corrupt or an old version. Overwriting.')
    if version > 1:
      raise NewerCredentialStoreError(
          'Credential file has file_version of %d. '
          'Only file_version of 1 is supported.' % version)

    credentials = []
    try:
      credentials = raw_data['data']
    except (TypeError, KeyError):
      pass

    for cred_entry in credentials:
      try:
        (key, credential) = self._decode_credential_from_json(cred_entry)
        self._data[key] = credential
      except:
        # If something goes wrong loading a credential, just ignore it
        logger.info('Error decoding credential, skipping', exc_info=True)

  def _decode_credential_from_json(self, cred_entry):
    """Load a credential from our JSON serialization.

    Args:
      cred_entry: A dict entry from the data member of our format

    Returns:
      (key, cred) where the key is the key tuple and the cred is the
        OAuth2Credential object.
    """
    raw_key = cred_entry['key']
    client_id = raw_key['clientId']
    user_agent = raw_key['userAgent']
    scope = raw_key['scope']
    key = (client_id, user_agent, scope)
    credential = None
    credential = Credentials.new_from_json(simplejson.dumps(cred_entry['credential']))
    return (key, credential)

  def _write(self):
    """Write the cached data back out.

    The multistore must be locked.
    """
    raw_data = {'file_version': 1}
    raw_creds = []
    raw_data['data'] = raw_creds
    for (cred_key, cred) in self._data.items():
      raw_key = {
          'clientId': cred_key[0],
          'userAgent': cred_key[1],
          'scope': cred_key[2]
          }
      raw_cred = simplejson.loads(cred.to_json())
      raw_creds.append({'key': raw_key, 'credential': raw_cred})
    self._locked_json_write(raw_data)

  def _get_credential(self, client_id, user_agent, scope):
    """Get a credential from the multistore.

    The multistore must be locked.

    Args:
      client_id: The client_id for the credential
      user_agent: The user agent for the credential
      scope: A string for the scope(s) being requested

    Returns:
      The credential specified or None if not present
    """
    key = (client_id, user_agent, scope)

    return self._data.get(key, None)

  def _update_credential(self, cred, scope):
    """Update a credential and write the multistore.

    This must be called when the multistore is locked.

    Args:
      cred: The OAuth2Credential to update/set
      scope: The scope(s) that this credential covers
    """
    key = (cred.client_id, cred.user_agent, scope)
    self._data[key] = cred
    self._write()

  def _delete_credential(self, client_id, user_agent, scope):
    """Delete a credential and write the multistore.

    This must be called when the multistore is locked.

    Args:
      client_id: The client_id for the credential
      user_agent: The user agent for the credential
      scope: The scope(s) that this credential covers
    """
    key = (client_id, user_agent, scope)
    try:
      del self._data[key]
    except KeyError:
      pass
    self._write()

  def _get_storage(self, client_id, user_agent, scope):
    """Get a Storage object to get/set a credential.

    This Storage is a 'view' into the multistore.

    Args:
      client_id: The client_id for the credential
      user_agent: The user agent for the credential
      scope: A string for the scope(s) being requested

    Returns:
      A Storage object that can be used to get/set this cred
    """
    return self._Storage(self, client_id, user_agent, scope)



#!/usr/bin/env python

import sys
import Personis
import Personis_base
import Personis_a
from Personis_util import printcomplist, printjson

# export a model sub tree to JSON

um = Personis_a.Access(model="Alice", modeldir='Tests/Models', user='alice', password='secret')
modeljson = um.export_model(["Personal"], resolver="default")
#modeljson = um.export_model(["Personal"])
printjson(modeljson)

um.import_model(context=["Temp"], partial_model=modeljson)
#um.import_model(partial_model=modeljson)

modeljson = um.export_model(["Temp"])
printjson(modeljson)



#!/bin/env python

# The Personis system is copyright 2000-2012 University of Sydney
#       Bob.Kummerfeld@Sydney.edu.au

# This file is part of Personis.

# Personis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Personis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Personis.  If not, see <http://www.gnu.org/licenses/>.

#
# mixin for Access class containing the evidence filters
#

import time,types,traceback,re
import json



class Access:
	"""Evidence filter mixin for Personis Access

	Contains evidence filter methods that are used before a resolver runs
	To add a new filter, add the method to this class and add an
	entry to the self.evidencefilterlist dictionary
	"""
	def __init__(self):
		self.evidencefilterlist = {"all":self._All_filter, "last1":self.last1_filter, "last10":self.last10_filter, "goal":self.goal_filter, "since_time":self.since_time_filter, "recent_count":self.recent_count_filter}


	def _All_filter(self, elist=None, arguments=None):
		"""
			return all evidence
		"""
		return sorted(elist, key=lambda e: e['creation_time'])
	
	def last1_filter(self, elist=None, arguments=None):
		"""
			return last 1 evidence item
		"""
		elist = sorted(elist, key=lambda e: e['creation_time'])
		return [elist[-1]]
	
	def last10_filter(self, elist=None, arguments=None):
		"""
			return last 10 evidence items
		"""
		elist = sorted(elist, key=lambda e: e['creation_time'])
		return elist[-1:-10]
	
	def goal_filter(self, elist=None, arguments=None):
		"""
			return elements with flag "goal"
		"""
		elist = [e for e in elist if "goal" in e['flags']]
		return sorted(elist, key=lambda e: e['creation_time'])

	def since_time_filter(self, elist=None, arguments=None):
		"""
			return elements with creation_time after arguments["since_time"]
		"""
		try:
			since_time = float(arguments['since_time'])
		except:
			since_time = 0.0
		elist = [e for e in elist if e['creation_time'] > since_time]
		return sorted(elist, key=lambda e: e['creation_time'])

	def recent_count_filter(self, elist=None, arguments=None):
		"""
			return the last arguments["recent_count"] elements
		"""
		elist = sorted(elist, key=lambda e: e['creation_time'])
		return elist[-1:-arguments['recent_count']]

class Evidence_list:
	def __init__(self, elist=None, model=None, componentid=None):
		self.content = elist
		if model != None:
			self.model = model
			self.componentid = componentid
			try:
				self.evdb,self.evdb_fd = shelf_open(model.curcontext+"/.evidence", "r")
			except:
				raise ValueError("tell: no evidence db for %s"%(model.curcontext))
			if self.componentid not in self.evdb:
				shelf_close(self.evdb, self.evdb_fd)
				return None # no evidence for this component
			self.totalcount = self.evdb[self.componentid]

	def close(self):
		shelf_close(self.evdb, self.evdb_fd)
		
	def _dbget(index):
		if index < 0:
			index = self.totalcount + index + 1
		return self.evdb.get("%s:%d"%(self.componentid,index))

	def __getitem__(self, index):
		if type(index) == type(1):
			if self.content == None:
				return Evidence_list([self._dbget(index)])
			else:
				if index < 0:
					index = self.totalcount + index + 1
				if (index < len(self.content)) and (index >= 0):
					return Evidence_list([self.content[index]])
				else:
					return Evidence_list([])
		elif type(index) == type(slice):
			step = index.step if index.step != None else 1
			elist = []
			for i in range(index.start, index.stop, step):
				if self.content == None:
					e = self._dbget(index)
					if e != None:
						elist.append(e)
				else:
					if index < 0:
						index = self.totalcount + index + 1
					if (index < len(self.content)) and (index >= 0):
						elist.append(self.content[index])
			return Evidence_list(elist)




#!/bin/env python

# The Personis system is copyright 2000-2012 University of Sydney
#       Bob.Kummerfeld@Sydney.edu.au

# This file is part of Personis.

# Personis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Personis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Personis.  If not, see <http://www.gnu.org/licenses/>.

#
# mixin for Access class containing the resolvers
#

import time,types,traceback,re
import json


class Access:
	"""Resolver mixin for Personis Access

	Contains resolver methods that can be specified on an "ask" operation
	To add a new resolver, add the method to this class and add an
	entry to the self.resolverlist dictionary
	"""
	def __init__(self):
		self.resolverlist = {"default":self._DefaultResolver, "goal":self.GoalResolver, "recent":self.RecentResolver}

	
	def _DefaultResolver(self, model=None, component=None, context=None, resolver_args=None):
		"""resolver function used if none specified
		sets component value to the value from the 
		last piece of given, non-goal evidence
		"""
		ev, evcount = component.getevidence(model=model, context=context)
		while evcount > 0:
			if ev != None:
				if not ("goal" in ev.flags):
					component.value = ev.value
					break
				else:
					ev, evcount = component.getevidence(model=model, context=context, count=evcount-1)
					continue
			else:
				component.value = None
				break
		if evcount == 0:
			component.value = None
		component.evidencelist = component.filterevidence(model=self, context=context, resolver_args=resolver_args)
		return component

	def GoalResolver(self, model=None, component=None, context=None, resolver_args=None):
		"""	resolver function used if none specified
		sets component value to the value from the 
		last piece of given, goal evidence
		"""
		ev, evcount = component.getevidence(model=model, context=context)
		while evcount > 0:
			if ev != None:
				if "goal" in ev.flags:
					component.value = ev.value
					break
				else:
					ev, evcount = component.getevidence(model=model, context=context, count=evcount-1)
					continue
			else:
				component.value = None
				break
		if evcount == 0:
			component.value = None
		component.evidencelist = component.filterevidence(model=self, context=context, resolver_args=resolver_args)
		return component

	def RecentResolver(self, model=None, component=None, context=None, resolver_args=None):
		"""	resolver function used if none specified
		sets component value to the value from the 
		last piece of given, goal evidence
		"""
		ev, evcount = component.getevidence(model=model, context=context)
		
		if resolver_args is None:
			return component
		component.evidencelist = component.filterevidence(model=self, context=context, resolver_args=resolver_args)
		return component




"""A fast, lightweight, and secure session WSGI middleware for use with GAE."""
from Cookie import CookieError, SimpleCookie
from base64 import b64decode, b64encode
import datetime
import hashlib
import hmac
import logging
import pickle
import os
import threading
import time

from google.appengine.api import memcache
from google.appengine.ext import db

# Configurable cookie options
COOKIE_NAME_PREFIX = "DgU"  # identifies a cookie as being one used by gae-sessions (so you can set cookies too)
COOKIE_PATH = "/"
DEFAULT_COOKIE_ONLY_THRESH = 10240  # 10KB: GAE only allows ~16000B in HTTP header - leave ~6KB for other info
DEFAULT_LIFETIME = datetime.timedelta(days=7)

# constants
SID_LEN = 43  # timestamp (10 chars) + underscore + md5 (32 hex chars)
SIG_LEN = 44  # base 64 encoded HMAC-SHA256
MAX_COOKIE_LEN = 4096
EXPIRE_COOKIE_FMT = ' %s=; expires=Wed, 01-Jan-1970 00:00:00 GMT; Path=' + COOKIE_PATH
COOKIE_FMT = ' ' + COOKIE_NAME_PREFIX + '%02d="%s"; %sPath=' + COOKIE_PATH + '; HttpOnly'
COOKIE_FMT_SECURE = COOKIE_FMT + '; Secure'
COOKIE_DATE_FMT = '%a, %d-%b-%Y %H:%M:%S GMT'
COOKIE_OVERHEAD = len(COOKIE_FMT % (0, '', '')) + len('expires=Xxx, xx XXX XXXX XX:XX:XX GMT; ') + 150  # 150=safety margin (e.g., in case browser uses 4000 instead of 4096)
MAX_DATA_PER_COOKIE = MAX_COOKIE_LEN - COOKIE_OVERHEAD

_tls = threading.local()
def get_current_session():
    """Returns the session associated with the current request."""
    return _tls.current_session

def set_current_session(session):
    """Sets the session associated with the current request."""
    _tls.current_session = session

def is_gaesessions_key(k):
    return k.startswith(COOKIE_NAME_PREFIX)

class SessionModel(db.Model):
    """Contains session data.  key_name is the session ID and pdump contains a
    pickled dictionary which maps session variables to their values."""
    pdump = db.BlobProperty()

class Session(object):
    """Manages loading, reading/writing key-value pairs, and saving of a session.

    ``sid`` - if set, then the session for that sid (if any) is loaded. Otherwise,
    sid will be loaded from the HTTP_COOKIE (if any).
    """
    DIRTY_BUT_DONT_PERSIST_TO_DB = 1

    def __init__(self, sid=None, lifetime=DEFAULT_LIFETIME, no_datastore=False,
                 cookie_only_threshold=DEFAULT_COOKIE_ONLY_THRESH, cookie_key=None):
        self._accessed = False
        self.sid = None
        self.cookie_keys = []
        self.cookie_data = None
        self.data = {}
        self.dirty = False  # has the session been changed?

        self.lifetime = lifetime
        self.no_datastore = no_datastore
        self.cookie_only_thresh = cookie_only_threshold
        self.base_key = cookie_key

        if sid:
            self.__set_sid(sid, False)
            self.data = None
        else:
            self.__read_cookie()

    @staticmethod
    def __compute_hmac(base_key, sid, text):
        """Computes the signature for text given base_key and sid."""
        key = base_key + sid
        return b64encode(hmac.new(key, text, hashlib.sha256).digest())

    def __read_cookie(self):
        """Reads the HTTP Cookie and loads the sid and data from it (if any)."""
        try:
            # check the cookie to see if a session has been started
            cookie = SimpleCookie(os.environ['HTTP_COOKIE'])
            self.cookie_keys = filter(is_gaesessions_key, cookie.keys())
            if not self.cookie_keys:
                return  # no session yet
            self.cookie_keys.sort()
            data = ''.join(cookie[k].value for k in self.cookie_keys)
            i = SIG_LEN + SID_LEN
            sig, sid, b64pdump = data[:SIG_LEN], data[SIG_LEN:i], data[i:]
            pdump = b64decode(b64pdump)
            actual_sig = Session.__compute_hmac(self.base_key, sid, pdump)
            if sig == actual_sig:
                self.__set_sid(sid, False)
                # check for expiration and terminate the session if it has expired
                if self.get_expiration() != 0 and time.time() > self.get_expiration():
                    return self.terminate()

                if pdump:
                    self.data = self.__decode_data(pdump)
                else:
                    self.data = None  # data is in memcache/db: load it on-demand
            else:
                logging.warn('cookie with invalid sig received from %s: %s' % (os.environ.get('REMOTE_ADDR'), b64pdump))
        except (CookieError, KeyError, IndexError, TypeError):
            # there is no cookie (i.e., no session) or the cookie is invalid
            self.terminate(False)

    def make_cookie_headers(self):
        """Returns a list of cookie headers to send (if any)."""
        # expire all cookies if the session has ended
        if not self.sid:
            return [EXPIRE_COOKIE_FMT % k for k in self.cookie_keys]

        if self.cookie_data is None:
            return []  # no cookie headers need to be sent

        # build the cookie header(s): includes sig, sid, and cookie_data
        if self.is_ssl_only():
            m = MAX_DATA_PER_COOKIE - 8
            fmt = COOKIE_FMT_SECURE
        else:
            m = MAX_DATA_PER_COOKIE
            fmt = COOKIE_FMT
        sig = Session.__compute_hmac(self.base_key, self.sid, self.cookie_data)
        cv = sig + self.sid + b64encode(self.cookie_data)
        num_cookies = 1 + (len(cv) - 1) / m
        if self.get_expiration() > 0:
            ed = "expires=%s; " % datetime.datetime.fromtimestamp(self.get_expiration()).strftime(COOKIE_DATE_FMT)
        else:
            ed = ''
        cookies = [fmt % (i, cv[i*m:i*m+m], ed) for i in xrange(num_cookies)]

        # expire old cookies which aren't needed anymore
        old_cookies = xrange(num_cookies, len(self.cookie_keys))
        key = COOKIE_NAME_PREFIX + '%02d'
        cookies_to_ax = [EXPIRE_COOKIE_FMT % (key % i) for i in old_cookies]
        return cookies + cookies_to_ax

    def is_active(self):
        """Returns True if this session is active (i.e., it has been assigned a
        session ID and will be or has been persisted)."""
        return self.sid is not None

    def is_ssl_only(self):
        """Returns True if cookies set by this session will include the "Secure"
        attribute so that the client will only send them over a secure channel
        like SSL)."""
        return self.sid is not None and self.sid[-33]=='S'

    def is_accessed(self):
        """Returns True if any value of this session has been accessed."""
        return self._accessed

    def ensure_data_loaded(self):
        """Fetch the session data if it hasn't been retrieved it yet."""
        self._accessed = True
        if self.data is None and self.sid:
            self.__retrieve_data()

    def get_expiration(self):
        """Returns the timestamp at which this session will expire."""
        try:
            return int(self.sid[:-33])
        except:
            return 0

    def __make_sid(self, expire_ts=None, ssl_only=False):
        """Returns a new session ID."""
        # make a random ID (random.randrange() is 10x faster but less secure?)
        if expire_ts is None:
            expire_dt = datetime.datetime.now() + self.lifetime
            expire_ts = int(time.mktime((expire_dt).timetuple()))
        else:
            expire_ts = int(expire_ts)
        if ssl_only:
            sep = 'S'
        else:
            sep = '_'
        return ('%010d' % expire_ts) + sep + hashlib.md5(os.urandom(16)).hexdigest()

    @staticmethod
    def __encode_data(d):
        """Returns a "pickled+" encoding of d.  d values of type db.Model are
        protobuf encoded before pickling to minimize CPU usage & data size."""
        # separate protobufs so we'll know how to decode (they are just strings)
        eP = {} # for models encoded as protobufs
        eO = {} # for everything else
        for k,v in d.iteritems():
            if isinstance(v, db.Model):
                eP[k] = db.model_to_protobuf(v)
            else:
                eO[k] = v
        return pickle.dumps((eP,eO), 2)

    @staticmethod
    def __decode_data(pdump):
        """Returns a data dictionary after decoding it from "pickled+" form."""
        try:
            eP, eO = pickle.loads(pdump)
            for k,v in eP.iteritems():
                eO[k] = db.model_from_protobuf(v)
        except Exception, e:
            logging.warn("failed to decode session data: %s" % e)
            eO = {}
        return eO

    def regenerate_id(self, expiration_ts=None):
        """Assigns the session a new session ID (data carries over).  This
        should be called whenever a user authenticates to prevent session
        fixation attacks.

        ``expiration_ts`` - The UNIX timestamp the session will expire at. If
        omitted, the session expiration time will not be changed.
        """
        if self.sid or expiration_ts is not None:
            self.ensure_data_loaded()  # ensure we have the data before we delete it
            if expiration_ts is None:
                expiration_ts = self.get_expiration()
            self.__set_sid(self.__make_sid(expiration_ts, self.is_ssl_only()))
            self.dirty = True  # ensure the data is written to the new session

    def start(self, expiration_ts=None, ssl_only=False):
        """Starts a new session.  expiration specifies when it will expire.  If
        expiration is not specified, then self.lifetime will used to
        determine the expiration date.

        Normally this method does not need to be called directly - a session is
        automatically started when the first value is added to the session.

        ``expiration_ts`` - The UNIX timestamp the session will expire at. If
        omitted, the session will expire after the default ``lifetime`` has past
        (as specified in ``SessionMiddleware``).

        ``ssl_only`` - Whether to specify the "Secure" attribute on the cookie
        so that the client will ONLY transfer the cookie over a secure channel.
        """
        self.dirty = True
        self.data = {}
        self.__set_sid(self.__make_sid(expiration_ts, ssl_only), True)

    def terminate(self, clear_data=True):
        """Deletes the session and its data, and expires the user's cookie."""
        if clear_data:
            self.__clear_data()
        self.sid = None
        self.data = {}
        self.dirty = False
        if self.cookie_keys:
            self.cookie_data = ''  # trigger the cookies to expire
        else:
            self.cookie_data = None

    def __set_sid(self, sid, make_cookie=True):
        """Sets the session ID, deleting the old session if one existed.  The
        session's data will remain intact (only the session ID changes)."""
        if self.sid:
            self.__clear_data()
        self.sid = sid
        self.db_key = db.Key.from_path(SessionModel.kind(), sid, namespace='')

        # set the cookie if requested
        if make_cookie:
            self.cookie_data = ''  # trigger the cookie to be sent

    def __clear_data(self):
        """Deletes this session from memcache and the datastore."""
        if self.sid:
            memcache.delete(self.sid, namespace='') # not really needed; it'll go away on its own
            try:
                db.delete(self.db_key)
            except:
                pass # either it wasn't in the db (maybe cookie/memcache-only) or db is down => cron will expire it

    def __retrieve_data(self):
        """Sets the data associated with this session after retrieving it from
        memcache or the datastore.  Assumes self.sid is set.  Checks for session
        expiration after getting the data."""
        pdump = memcache.get(self.sid, namespace='')
        if pdump is None:
            # memcache lost it, go to the datastore
            if self.no_datastore:
                logging.info("can't find session data in memcache for sid=%s (using memcache only sessions)" % self.sid)
                self.terminate(False) # we lost it; just kill the session
                return
            session_model_instance = db.get(self.db_key)
            if session_model_instance:
                pdump = session_model_instance.pdump
            else:
                logging.error("can't find session data in the datastore for sid=%s" % self.sid)
                self.terminate(False) # we lost it; just kill the session
                return
        self.data = self.__decode_data(pdump)

    def save(self, persist_even_if_using_cookie=False):
        """Saves the data associated with this session IF any changes have been
        made (specifically, if any mutator methods like __setitem__ or the like
        is called).

        If the data is small enough it will be sent back to the user in a cookie
        instead of using memcache and the datastore.  If `persist_even_if_using_cookie`
        evaluates to True, memcache and the datastore will also be used.  If the
        no_datastore option is set, then the datastore will never be used.

        Normally this method does not need to be called directly - a session is
        automatically saved at the end of the request if any changes were made.
        """
        if not self.sid:
            return # no session is active
        if not self.dirty:
            return # nothing has changed
        dirty = self.dirty
        self.dirty = False  # saving, so it won't be dirty anymore

        # do the pickling ourselves b/c we need it for the datastore anyway
        pdump = self.__encode_data(self.data)

        # persist via cookies if it is reasonably small
        if len(pdump)*4/3 <= self.cookie_only_thresh: # 4/3 b/c base64 is ~33% bigger
            self.cookie_data = pdump
            if not persist_even_if_using_cookie:
                return
        elif self.cookie_keys:
            # latest data will only be in the backend, so expire data cookies we set
            self.cookie_data = ''

        memcache.set(self.sid, pdump, namespace='', time=self.get_expiration())  # may fail if memcache is down

        # persist the session to the datastore
        if dirty is Session.DIRTY_BUT_DONT_PERSIST_TO_DB or self.no_datastore:
            return
        try:
            SessionModel(key_name=self.sid, pdump=pdump).put()
        except Exception, e:
            logging.warning("unable to persist session to datastore for sid=%s (%s)" % (self.sid,e))

    # Users may interact with the session through a dictionary-like interface.
    def clear(self):
        """Removes all data from the session (but does not terminate it)."""
        if self.sid:
            self.data = {}
            self.dirty = True

    def get(self, key, default=None):
        """Retrieves a value from the session."""
        self.ensure_data_loaded()
        return self.data.get(key, default)

    def has_key(self, key):
        """Returns True if key is set."""
        self.ensure_data_loaded()
        return self.data.has_key(key)

    def pop(self, key, default=None):
        """Removes key and returns its value, or default if key is not present."""
        self.ensure_data_loaded()
        self.dirty = True
        return self.data.pop(key, default)

    def pop_quick(self, key, default=None):
        """Removes key and returns its value, or default if key is not present.
        The change will only be persisted to memcache until another change
        necessitates a write to the datastore."""
        self.ensure_data_loaded()
        if self.dirty is False:
            self.dirty = Session.DIRTY_BUT_DONT_PERSIST_TO_DB
        return self.data.pop(key, default)

    def set_quick(self, key, value):
        """Set a value named key on this session.  The change will only be
        persisted to memcache until another change necessitates a write to the
        datastore.  This will start a session if one is not already active."""
        dirty = self.dirty
        self[key] = value
        if dirty is False or dirty is Session.DIRTY_BUT_DONT_PERSIST_TO_DB:
            self.dirty = Session.DIRTY_BUT_DONT_PERSIST_TO_DB

    def __getitem__(self, key):
        """Returns the value associated with key on this session."""
        self.ensure_data_loaded()
        return self.data.__getitem__(key)

    def __setitem__(self, key, value):
        """Set a value named key on this session.  This will start a session if
        one is not already active."""
        self.ensure_data_loaded()
        if not self.sid:
            self.start()
        self.data.__setitem__(key, value)
        self.dirty = True

    def __delitem__(self, key):
        """Deletes the value associated with key on this session."""
        self.ensure_data_loaded()
        self.data.__delitem__(key)
        self.dirty = True

    def __iter__(self):
        """Returns an iterator over the keys (names) of the stored values."""
        self.ensure_data_loaded()
        return self.data.iterkeys()

    def __contains__(self, key):
        """Returns True if key is present on this session."""
        self.ensure_data_loaded()
        return self.data.__contains__(key)

    def __str__(self):
        """Returns a string representation of the session."""
        if self.sid:
            self.ensure_data_loaded()
            return "SID=%s %s" % (self.sid, self.data)
        else:
            return "uninitialized session"

class SessionMiddleware(object):
    """WSGI middleware that adds session support.

    ``cookie_key`` - A key used to secure cookies so users cannot modify their
    content.  Keys should be at least 32 bytes (RFC2104).  Tip: generate your
    key using ``os.urandom(64)`` but do this OFFLINE and copy/paste the output
    into a string which you pass in as ``cookie_key``.  If you use ``os.urandom()``
    to dynamically generate your key at runtime then any existing sessions will
    become junk every time your app starts up!

    ``lifetime`` - ``datetime.timedelta`` that specifies how long a session may last.  Defaults to 7 days.

    ``no_datastore`` - By default all writes also go to the datastore in case
    memcache is lost.  Set to True to never use the datastore. This improves
    write performance but sessions may be occassionally lost.

    ``cookie_only_threshold`` - A size in bytes.  If session data is less than this
    threshold, then session data is kept only in a secure cookie.  This avoids
    memcache/datastore latency which is critical for small sessions.  Larger
    sessions are kept in memcache+datastore instead.  Defaults to 10KB.
    """
    def __init__(self, app, cookie_key, lifetime=DEFAULT_LIFETIME, no_datastore=False, cookie_only_threshold=DEFAULT_COOKIE_ONLY_THRESH):
        self.app = app
        self.lifetime = lifetime
        self.no_datastore = no_datastore
        self.cookie_only_thresh = cookie_only_threshold
        self.cookie_key = cookie_key
        if not self.cookie_key:
            raise ValueError("cookie_key MUST be specified")
        if len(self.cookie_key) < 32:
            raise ValueError("RFC2104 recommends you use at least a 32 character key.  Try os.urandom(64) to make a key.")

    def __call__(self, environ, start_response):
        # initialize a session for the current user
        _tls.current_session = Session(lifetime=self.lifetime, no_datastore=self.no_datastore, cookie_only_threshold=self.cookie_only_thresh, cookie_key=self.cookie_key)

        # create a hook for us to insert a cookie into the response headers
        def my_start_response(status, headers, exc_info=None):
            _tls.current_session.save() # store the session if it was changed
            for ch in _tls.current_session.make_cookie_headers():
                headers.append(('Set-Cookie', ch))
            return start_response(status, headers, exc_info)

        # let the app do its thing
        return self.app(environ, my_start_response)

class DjangoSessionMiddleware(object):
    """Django middleware that adds session support.  You must specify the
    session configuration parameters by modifying the call to ``SessionMiddleware``
    in ``DjangoSessionMiddleware.__init__()`` since Django cannot call an
    initialization method with parameters.
    """
    def __init__(self):
        fake_app = lambda environ, start_response : start_response
        self.wrapped_wsgi_middleware = SessionMiddleware(fake_app, cookie_key='you MUST change this')
        self.response_handler = None

    def process_request(self, request):
        self.response_handler = self.wrapped_wsgi_middleware(None, lambda status, headers, exc_info : headers)
        request.session = get_current_session()  # for convenience

    def process_response(self, request, response):
        if self.response_handler:
            session_headers = self.response_handler(None, [], None)
            for k,v in session_headers:
                response[k] = v
            self.response_handler = None
        if request.session.is_accessed():
            from django.utils.cache import patch_vary_headers
            logging.info("Varying")
            patch_vary_headers(response, ('Cookie',))
        return response

def delete_expired_sessions():
    """Deletes expired sessions from the datastore.
    If there are more than 500 expired sessions, only 500 will be removed.
    Returns True if all expired sessions have been removed.
    """
    now_str = unicode(int(time.time()))
    q = db.Query(SessionModel, keys_only=True, namespace='')
    key = db.Key.from_path('SessionModel', now_str + u'\ufffd', namespace='')
    q.filter('__key__ < ', key)
    results = q.fetch(500)
    db.delete(results)
    logging.info('gae-sessions: deleted %d expired sessions from the datastore' % len(results))
    return len(results)<500



#!/usr/bin/env python

import sys

import os
import platform
import time
from personis import client
import json

from oauth2client.file import Storage
from oauth2client.client import Credentials, OAuth2WebServerFlow, flow_from_clientsecrets
from oauth2client.tools import run
import httplib2
import logging
from Queue import Queue
import threading
import math

osname = os.name + platform.system()
if osname == 'posixLinux':
    import ctypes
    class XScreenSaverInfo( ctypes.Structure):
        """ typedef struct { ... } XScreenSaverInfo; """
        _fields_ = [('window',      ctypes.c_ulong), # screen saver window
                    ('state',       ctypes.c_int),   # off,on,disabled
                    ('kind',        ctypes.c_int),   # blanked,internal,external
                    ('since',       ctypes.c_ulong), # milliseconds
                    ('idle',        ctypes.c_ulong), # milliseconds
                    ('event_mask',  ctypes.c_ulong)] # events

    class idler(object):
        def __init__(self):
            self.xlib = ctypes.cdll.LoadLibrary( 'libX11.so')
            #display_num = os.environ['DISPLAY']
            display_num = ":0"
            self.dpy = self.xlib.XOpenDisplay(display_num)
            if not self.dpy:
                raise Exception('Cannot open display')
            self.root = self.xlib.XDefaultRootWindow( self.dpy)
            self.xss = ctypes.cdll.LoadLibrary( 'libXss.so.1')
            self.xss.XScreenSaverAllocInfo.restype = ctypes.POINTER(XScreenSaverInfo)
            self.xss_info = self.xss.XScreenSaverAllocInfo()
            self.current = 0

        def getIdleTime(self):
            self.xss.XScreenSaverQueryInfo( self.dpy, self.root, self.xss_info)
            return int(self.xss_info.contents.idle / 1000)

elif osname == 'ntWindows':
    import win32api
    class idler(object):
        def __init__(self):
            self.current = 0
        def getIdleTime(self):
            return win32api.GetLastInputInfo()/1000

elif osname == 'posixDarwin':
    from Quartz.CoreGraphics import CGEventSourceSecondsSinceLastEventType
    # From /System/Library/Frameworks/IOKit.framework/Versions/A/Headers/hidsystem/IOLLEvent.h
    NX_ALLEVENTS = int(4294967295)  # 32-bits, all on.
    class idler(object):
        def getIdleTime(self):
            return CGEventSourceSecondsSinceLastEventType(1, NX_ALLEVENTS)
else:
    raise Error()


def install_inactivity(um):
    try:
        reslist = um.ask(context=["Devices","activity_monitor"])
        return
    except:
        pass

    ctx_obj = client.Context(Identifier="activity_monitor", Description="Watch computer usage",
                 perms={'ask':True, 'tell':True, "resolvers": ["all","last10","last1","goal"]},
                 resolver=None, objectType="Context")
    context = ['Devices']
    um.mkcontext(context,ctx_obj)
    context.append('activity_monitor')

    ctx_obj = client.Context(Identifier="activity", Description="Extract data from activity watcher",
                 perms={'ask':True, 'tell':True, "resolvers":["all","last10","last1","goal"]},
                 resolver=None, objectType="Context")
    um.mkcontext(context,ctx_obj)
    context.append('activity')

    cobj = client.Component(Identifier="data", component_type="attribute", 
        value_type="number",resolver=None,Description="0 on inactivity detection, 1 on activity detection, -1 on shutdown")
    um.mkcomponent(context=context, componentobj=cobj)

class sender(threading.Thread):
    def __init__(self, um):
        self.q = Queue()
        self.um = um
        self.running = False
        threading.Thread.__init__(self)

    def send(self, v, t):
        ev = client.Evidence(source='activity_monitor', evidence_type="explicit", value=v, time=t)
        self.q.put(ev)
        logging.debug('put %s'%(ev))

    def run(self):
        retries = 1
        self.running = True
        ev = None
        while self.running:
            if ev == None:
                ev = self.q.get(block=True)
                logging.debug('got %s'%(ev))
            try:
                self.um.tell(context=['Devices','activity_monitor','activity'], componentid='data', evidence=ev)
                logging.info('%s: %s'%(time.asctime(time.localtime(ev.time)), ev.value))
                ev = None
                retries = 1
                self.q.task_done()
            except Exception as e:
                backoff = math.sqrt(retries)*4
                logging.info('Personis down? try again in %d seconds'%(backoff))
                time.sleep(backoff)
                retries = retries + 1

    def stop(self):
        self.running = False
        self.q.join()
        threading.Thread.join(self)



inactive_granularity = 5 # seconds
active_granularity = 10 # seconds - you can stop typing for 10 seconds and it's not seen.

#httplib2.debuglevel=0
#logging.basicConfig(level=logging.INFO)

# get past the uni's stupid proxy server
p = httplib2.ProxyInfo(proxy_type=httplib2.socks.PROXY_TYPE_HTTP_NO_TUNNEL, proxy_host='www-cache.it.usyd.edu.au', proxy_port=8000)

# Use the util package to get a link to UM. This uses the client_secrets.json file for the um location
client_secrets = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'client_secrets.json')
um = client.util.LoginFromClientSecrets(filename=client_secrets, http=httplib2.Http(proxy_info=p, disable_ssl_certificate_validation=True), credentials='activity_cred.dat')

reslist = um.ask(context=["Personal"],view=['firstname'])
#print 'logging for', reslist[0].value

install_inactivity(um)
snd = sender(um)
snd.start()
idle = idler();
time.sleep(1)

idle_state = True
i = inactive_granularity + active_granularity
snd.send(0, time.time())                

try:
    while True:
            if idle_state:
                while i >= inactive_granularity:
                    time.sleep(inactive_granularity)
                    i = idle.getIdleTime()
                snd.send(1, time.time())                
                idle_state = False
            if not idle_state:
                while i < active_granularity:
                    time.sleep(active_granularity)
                    i = idle.getIdleTime()
                snd.send(0, time.time())                
                idle_state = True
except:
    pass
finally:
    snd.send(-1, time.time())  
    snd.stop()


#!/usr/local/bin/python

import Personis_base
from Personis_util import printcomplist

print("tell new location for Alice")
um = Personis_base.Access(model='Alice', modeldir='Tests/Models', user='alice', password='secret')
# create a piece of evidence with home as value
ev = Personis_base.Evidence(evidence_type="explicit", value="home")
# tell this as user alice's location
um.tell(context=['Location'], componentid='location', evidence=ev)

print("===================================================================")
print("Now check the evidence list for Alice's location")
reslist = um.ask(context=['Location'], view=['location'], resolver={'evidence_filter':"all"})
printcomplist(reslist, printev = "yes")




#!/usr/bin/env python

import Personis
import Personis_base
from Personis_util import printcomplist

print(">>>> add some evidence to some components and then use example1 to see it")

um = Personis.Access(model="Alice", user='alice', password='secret')

print(">>>> create a piece of evidence with alice as source and Alice as value")
ev = Personis_base.Evidence(source="willy", evidence_type="explicit", value="Fred")
print(">>>> tell this as user alice's first name")
um.tell(context=["Personal"], componentid='firstname', evidence=ev)

ev = Personis_base.Evidence(source="nilly", evidence_type="explicit", value="Smith")
um.tell(context=["Personal"], componentid='lastname', evidence=ev)

print("===================================================================")
print(">>>> Now check the evidence list for Alice's names")
reslist = um.ask(context=["Personal"], view='fullname')
printcomplist(reslist, printev = "yes")




#!/usr/bin/env python

# The Personis system is copyright 2000-2012 University of Sydney
#       Bob.Kummerfeld@Sydney.edu.au

# This file is part of Personis.

# Personis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Personis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Personis.  If not, see <http://www.gnu.org/licenses/>.

"""
Personis_mkmodel - library functions used to create sets of models.
	Used by the program mkmodel

mkmodel takes a definition of a model (stored in modeldefinitionfile)
and creates a model in modeldirectory for each model 


"""

import sys
from . import base
from . import active
import logging


"""
	Functions for parsing modeldefs. Uses the pyparsing module 

 Modeldef statement grammar:

	name := ID
	flag := '[' QUOTEDSTRING [',' QUOTEDSTRING]* ']'
	value := QUOTEDSTRING | flags
	key_value := name '=' value
	evidence := '[' key_value* ']'
	attribute := key_value | evidence
	path = [ '/' ] name [ '/' name ]*
	context := '@@' path ':' attribute [ ',' attribute ]*
	component := '--' name ':' attribute [ ',' attribute ]*
	view := '==' name ':' path [ ',' path ]

 Example:

@@Location: description="Information about the user's location."
--seenby: type="attribute", value_type="string", description="sensor that has seen this person"
--location: type="attribute", value_type="string", description="Location Component", 
	[evidence_type="explicit", value="Bob"], [evidence_type="explicit", flags= [ "f1", "f2" ], value="blah"]
==fullname: Personal/firstname, Personal/lastname

"""
from .mypyparsing import Word, alphas, alphanums, Literal, ZeroOrMore, quotedString, Forward, removeQuotes, ParseException, Optional, OneOrMore

keyvals = {} # dictionary to hold list of key/value pairs for an element of the  modeldef
attrs = []
paths = [] # list of paths
flags = None
curcontext = ""
themodel = None
Debug = False


def doname(str, loc, toks):
#	print "doname::", toks
	pass

def doflaglist(str, loc, toks):
	global flags
#	print "doflaglist::", toks
	pass
	flags = []
	if (len(toks) < 3) or  (toks[0] != '[') or (toks[-1] != ']'):
		logging.debug("*** bad flag list: %s", toks)
		return
	for i in range(1,len(toks),2):
		flags.append(toks[i])
	ingging.info( "\tflaglist:", flags)

def dokeyval(str, loc, toks):
	global keyvals, flags
#	print "dokeyval::", toks
	if flags != None:
		val = flags
		flags = None
	else:
		val = toks[2]
	if toks[0] in keyvals:
		if type(keyvals[toks[0]]) == type([]):
			keyvals[toks[0]].append(val)
		else:
			keyvals[toks[0]] = [keyvals[toks[0]], val]
	else:
		keyvals[toks[0]] = val
#	print "\tkeyvals::", keyvals

def dokeyval_list(str, loc, toks):
	global keyvals, attrs
#	print "dokeyval_list::", toks
#	print "\tkeyvals", keyvals
	attrs.append(keyvals)
#	print " \tattrs::", attrs
	keyvals = {}


def doevidence(str, loc, toks):
	global attrs
#	print "doevidence:: ", toks
#	print "\t", attrs
	pass

def doevidencelist(str, loc, toks):
#	print "doevidencelist:: ", toks
	pass

def dotells(ev, compid):
	global themodel, curcontext
	evattrs = ["flags", "evidence_type", "source", "owner", "value", "comment", "time", "useby"]
	if not all([a in evattrs for a in ev]):
		logging.debug( "**** evidence attributes %s must be one of %s" % (ev.keys(), repr(evattrs)))
		return 
	if "flags" in ev:
		if type(ev['flags']) != type([]):
			logging.debug( "**** evidence flags %s must be a list" % (ev['flags']))
			return

	if not Debug:
		evobj = base.Evidence(evidence_type="explicit")
		for k,v in v.items():
			evobj.__dict__[k] = v
		themodel.tell(context=curcontext, componentid=compid, evidence=evobj)
		logging.debug(	"""
			evobj = base.Evidence(ev)
			themodel.tell(context=%s, componentid=%s, evidence=%s)
			""" % (curcontext, compid, evobj.__dict__))
	else:
		logging.debug(	"""
			evobj = base.Evidence(ev)
			themodel.tell(context=%s, componentid=%s, evidence=%s)
			""" % (curcontext, compid, ev))
		
		


def docomponent(str, loc, toks):
	global attrs, keyvals, paths, curcontext, themodel, Debug
	if curcontext == "":
		logging.debug( "No context defined for component", toks[1])
	logging.debug( "docomponent:: %s", toks[1])
	logging.debug( " \tattrs:: %s", attrs)
	required = ['type', 'description', 'value_type']
	for x in required:
		if x not in attrs[0]:
			logging.info( "one or more of the required keyvals", required, "not found for %s", toks[1])
			logging.info( attrs[0])
			return
	if not Debug:
		cobj = base.Component(Identifier=toks[1],
			component_type=attrs[0]['type'],
			value_type=attrs[0]['value_type'],
			value_list=attrs[0].get('value'),
			resolver=attrs[0].get('resolver'),
			Description=attrs[0]['description'])
		try:
			res = themodel.mkcomponent(context=curcontext, componentobj=cobj)
		except:
			logging.info( "mkcomponent failed")
		if res != None:
			logging.debug( res)
	logging.debug( """cobj = base.Component(Identifier="%s",
		component_type="%s",
		value_type="%s",
		value_list="%s",
		resolver="%s",
		Description="%s")
		""" % ( toks[1], attrs[0]['type'], attrs[0]['value_type'], attrs[0].get('value'), attrs[0].get('resolver'), attrs[0]['description']))
	if 'rule' in attrs[0]:
		if type(attrs[0]['rule']) != type([]):
			rules = [attrs[0]['rule']]
		for rule in rules:
			if not Debug:
				themodel.subscribe(context=curcontext, view=[toks[1]], subscription=dict(user="bob", password="qwert", statement=rule))
			logging.debug( "\tsub:: %s, %s, %s", curcontext, [toks[1]], dict(user="bob", password="qwert", statement=rule))
	logging.debug( "+++ component created ")
	if len(attrs) > 1: # see if there is some evidence
		for e in attrs[1:]:
			logging.debug( "\tevidence:: %s", e)
			dotells(e, toks[1])
#	del attrs[0]
	attrs = []
	keyvals = {}
	paths = []

def docontext(str, loc, toks):
	global attrs, paths, curcontext, themodel, Debug, keyvals
	logging.debug( "docontext:: %s", toks)
	logging.debug( " \tpaths:: %s", paths)
	logging.debug( " \tattrs:: %s", attrs)
	if len(paths) != 1:
		logging.info( "too many paths", paths)
		raise ParseException("too many paths " + repr(paths))
	curcontext = paths[0]
	logging.debug( "\tcurcontext:: "+ curcontext + repr(curcontext.split('/')))
	if 'description' not in attrs[0]:
		logging.info( "*** description required for ", curcontext)
		raise ParseException("description required for " + repr(curcontext))
	if not Debug:
		cobj = base.Context(Identifier=curcontext.split('/')[-1], Description=attrs[0]['description'])
	logging.debug( "\tbase.Context(Identifier='%s', Description='%s')", curcontext.split('/')[-1], attrs[0]['description'])
	logging.debug( "\t %s", curcontext.split('/')[:-1])
	if not Debug:
		if themodel.mkcontext(curcontext.split('/')[:-1], cobj):
			logging.debug( "+++ context created ok")
		else:
			logging.info( "+++ context creation failed")
	keyvals = {}
	del attrs[0]
	paths = []

def domdef(str, loc, toks):
	#print "domdef::", toks
	logging.debug( "--------------------------------")

def dopath(str, loc, toks):
	global paths
#	print "dopath::", toks
	paths.append(''.join(toks))

def doview(str, loc, toks):
	global paths, curcontext, themodel, Debug
	if curcontext == "":
		logging.info( "No context defined for view %s", toks[1])
		raise ParseException("No context defined for view " + repr(toks[1]))
	if paths == []:
		logging.info( "No paths defined for view %s", toks[1])
		raise ParseException("No paths defined for view " + repr(toks[1]))
	logging.debug( "doview:: %s", toks[1])
	logging.debug( "\t paths:: %s", paths)
	if not Debug:
		vobj = base.View(Identifier=toks[1], component_list=paths)
		themodel.mkview(curcontext, vobj)
	paths = []

#	name := ID
#	attribute := name '=' QUOTEDSTRING
#	path = [ '/' ] name [ '/' name ]
#	context := path ':' attribute [ ',' attribute ]*
#	component := name ':' attribute [ ',' attribute ]*
#	view := name ':' path [ ',' path ]

name = Word(alphanums+"_")
flaglist = Literal('[')+ quotedString.setParseAction(removeQuotes) + ZeroOrMore(Literal(',') + quotedString.setParseAction(removeQuotes)) + Literal(']')
value = quotedString.setParseAction(removeQuotes) | flaglist
keyval = name + Literal("=") + value
keyval_list = keyval + ZeroOrMore(Literal(",") + keyval)
evidence = Literal("[") + keyval_list + Literal("]")
evidencelist = evidence + ZeroOrMore(Literal(",") + evidence)
path = ZeroOrMore(Literal('/')) + name + ZeroOrMore(Literal('/') + name)
context = Literal('@@') + path + Literal(':') + keyval_list
component = Literal('--') + name + Literal(":") + keyval_list + Optional(Literal(',') + evidencelist)
view = Literal('==') + name + Literal(":") + path + ZeroOrMore(Literal(",") + path)
mdef = OneOrMore(context + ZeroOrMore(component) + ZeroOrMore(view)) + Literal('$$')

name.setParseAction(doname)
keyval.setParseAction(dokeyval)
flaglist.setParseAction(doflaglist)
keyval_list.setParseAction(dokeyval_list)
evidence.setParseAction(doevidence)
evidencelist.setParseAction(doevidencelist)
path.setParseAction(dopath)
context.setParseAction(docontext)
component.setParseAction(docomponent)
view.setParseAction(doview)
mdef.setParseAction(domdef)


def domodeldef(mdefstring):
	"""
	function to parse a modeldef statement
	arg is a string containing the mdef statement
	"""
	mdefstring += " $$"
	#print "statement:", mdefstring
	try:
		toks = mdef.parseString(mdefstring)
	except ParseException as err:
		logging.info( '****  Parse Failure  ****')
		logging.info( err.line)
		logging.info( " "*(err.column-1) + "^")
		logging.info( err)

		raise ValueError("parse failed")


"""====================================================================================================================="""
def mkmodel_um(um,lines, debug = 1):
	'''
		Create a model from the model definition in the string "lines"
	'''
	global themodel
	themodel = um
	domodeldef(lines)

#def mkmodel_remote(model=None, mfile=None, modelserver=None, user=None, password=None):
#	server.MkModel(model=model, modelserver=modelserver, user=user, password=password)
#	um = client.Access(model=model, modelserver=modelserver, user=user, password=password)
#	mkmodel_um(um,get_modeldef(mfile))
	

def mkmodel(model=None, mfile=None, modeldir=None, user=None, password=None):
	base.MkModel(model=model, modeldir=modeldir, user=user)
	um = active.Access(model=model, modeldir=modeldir, user=user)
	mkmodel_um(um,get_modeldef(mfile))

def get_modeldef(mfile):
	try:
		mf = open(mfile)
	except:
		logging.info( "cannot open <%s>" % (mfile))
		sys.exit(1)
	lines = ""
	for mline in mf.readlines():
#		mline = mline.strip()
#		if debug:
#			print ">"+mline
		if (len(mline) == 0) or (mline[0] in "#\n"):
			continue
		if mline[:9] == "$include ":
			inclfile = mline[9:].strip()
			logging.debug( "#### include file: %s\n" % (inclfile))
			lines = lines + get_modeldef(inclfile)
			logging.debug( "#### end of include file: %s\n" % (inclfile))
		else:
			lines = lines+mline
	mf.close()
	return lines


if __name__ == '__main__':
	Debug = True
	mdefstring = """
@@Location: description="Information about the user's location."
--seenby: type="attribute", value_type="string", description="sensor that has seen this person"
--location: type="attribute", value_type="string", description="Location Component", 
	[evidence_type="explicit", value="Bob"], [evidence_type="explicit", flags= [ "f1", "f2" ], value="blah"]
"""
	mmdefstring = """
@@Location: description="Information about the users' location."
--seenby: type="attribute", value_type="string", description="sensor that has seen this person"
--location: type="attribute", value_type="string", description="Location", [aa="hhh", bb="kk"]
==fullname: Personal/firstname, Personal/lastname
@@Personal/Work: description="Information about the users work."
--role: description="the users main role in the organisation", type="attribute",
        value_type="enum", value="Academic", value="Postgraduate", value="etc",
	rule="<default!./Personal/email> ~ '*' : NOTIFY 'http://www.somewhere.com/' 'email=' <./Personal/email>",
[aa="hhh", bb="jj"],
[aa="hhh", bb="kk"]
"""
	mmdefstring = """
@@Location: description="Information about the users' location."
--seenby: type="attribute", value_type="string", description="sensor that has seen this person"
--location: type="attribute", value_type="string", description="Location"
@@Work: description="Information about the users work."
--role: description="the users main role in the organisation", type="attribute", 
        value_type="enum", value="Academic", value="Postgraduate", value="etc"

"""
	mmdefstring = get_modeldef("modeldefs/user-test")
	mdefstring = get_modeldef(sys.argv[1])
	logging.debug( "=====================\n",mdefstring,"\n=====================\n")
#	domodeldef(mdefstring)




#!/usr/bin/env python

import Personis
import Personis_base
from Personis_util import showobj, printcomplist

def printAskContext( info ):
	(cobjlist, contexts, theviews, thesubs) = info
	printcomplist(cobjlist, printev = "yes")
	print("Contexts: %s" % str(contexts))
	print("Views: %s" % str(theviews))
	print("Subscriptions: %s" % str(thesubs))
		
print("===================================================================")
print("Examples that show how permissions work")
print("===================================================================")

um = Personis.Access(model="Alice", user='alice', password='secret')

print(">>>> Set some permissions for the 'MyHealth' app")
um.setpermission(context=["Personal"], app="MyHealth", permissions={'ask':True, 'tell':False})

print(">>>> Show the permissions")
perms = um.getpermission(context=["Personal"], app="MyHealth")
print("MyHealth:", perms)
perms = um.getpermission(context=["Personal"], app="withings")
print("withings:", perms)




#!/usr/bin/env python

import unittest
from personis.server import base, mkmodel
import os
import shutil
import sys
import logging

class TestPersonisBase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(level=logging.DEBUG)
        if os.path.exists('models'):
            shutil.rmtree('models')
        os.mkdir('models')

        mkmodel.mkmodel(model='alice', mfile='../modeldefs/user.mdef', modeldir='models', user='alice')
        cls.um = base.Access(model="alice", modeldir='models', user='alice')
        # create a piece of evidence with Alice as value
        ev = base.Evidence(evidence_type="explicit", value="Alice")
        # tell this as user alice's first name
        cls.um.tell(context=["Personal"], componentid='firstname', evidence=ev)

        ev = base.Evidence(evidence_type="explicit", value="Smith")
        cls.um.tell(context=["Personal"], componentid='lastname', evidence=ev)
        ev = base.Evidence(evidence_type="explicit", value="female")
        cls.um.tell(context=["Personal"], componentid='gender', evidence=ev)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('models')

    def test_ask_firstname(self):
        res = self.um.ask(context=["Personal"], view=['firstname'])
        self.assertEqual(res[0].value, 'Alice')

    def test_ask_gender(self):
        res = self.um.ask(context=["Personal"], view=['gender'], resolver={'evidence_filter':"last1"})
        self.assertEqual(res[0].value, 'female')

    def test_ask_fullname(self):
        res = self.um.ask(context=["Personal"], view='fullname')
        self.assertEqual(res[0].value, 'Alice')
        self.assertEqual(res[1].value, 'Smith')

    def test_ask_fullname_resolver(self):
        res = self.um.ask(context=["Personal"], view='fullname', resolver={'evidence_filter':"last1"})
        self.assertEqual(res[0].value, 'Alice')
        self.assertEqual(res[1].value, 'Smith')

    def test_ask_comboview(self):
        res = self.um.ask(context=["Personal"], view=['firstname', 'lastname'])
        self.assertEqual(res[0].value, 'Alice')
        self.assertEqual(res[1].value, 'Smith')

    def test_ask_more(self):
        res = self.um.ask(context=["Preferences", "Music", "Jazz", "Artists"], 
                        view=['Miles_Davis', ['Personal', 'firstname']],
                        resolver={'evidence_filter':"all"})
        self.assertEqual(res[1].Identifier, 'firstname')

if __name__ == '__main__':
    unittest.main()
